import logging
from collections import OrderedDict

import numpy as np
import xarray as xr

from . import resample, mgrid
from ..read import jarkus, rws
from .. import util

logger = logging.getLogger(__name__)


def calculate_thickness(ds, top="top", bot="bot"):
    """Calculate thickness from dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        dataset containing information about top and bottom elevations
        of layers
    top : str, optional
        name of data variable containing tops, by default "top"
    bot : str, optional
        name of data variable containing bottoms, by default "bot"

    Returns
    -------
    thickness : xarray.DataArray
        DataArray containing thickness information
    top3d : bool
        boolean whether each layer has a top elevation (True),
        or top only indicates surface level and bottom of layer i
        is top of layer i+1 (False)
    """
    # calculate thickness
    if ds[top].ndim == ds[bot].ndim and ds[top].ndim in [2, 3]:
        if ds[top].shape[0] == ds[bot].shape[0]:
            # top is 3D, every layer has top and bot
            thickness = ds[top] - ds[bot]
            top3d = True
        else:
            raise ValueError(
                '3d top and bot should have same number of layers')
    elif ds[top].ndim == (ds[bot].ndim - 1) and ds[top].ndim in [1, 2]:
        if ds[top].shape[-1] == ds[bot].shape[-1]:
            # top is only top of first layer
            thickness = xr.zeros_like(ds[bot])
            for lay in range(len(bot)):
                if lay == 0:
                    thickness[lay] = ds[top] - ds[bot][lay]
                else:
                    thickness[lay] = ds[bot][lay - 1] - ds[bot][lay]
            top3d = False
        else:
            raise ValueError('2d top should have same last dimension as bot')
    if isinstance(ds[bot], xr.DataArray):
        if hasattr(ds[bot], 'units'):
            if ds[bot].units == 'mNAP':
                thickness.attrs['units'] = 'm'
            else:
                thickness.attrs['units'] = ds[bot].units

    return thickness, top3d


def layer_split_top_bot(ds, split_dict, layer='layer', top='top', bot='bot'):
    """Calculate new tops and bottoms for split layers.

    Parameters
    ----------
    ds : xarray.Dataset
        xarray Dataset containing information about layers
        (layers, top and bot)
    split_dict : dict
        dictionary with index of layers to split as keys and iterable
        of fractions that add up to 1 to indicate how to split up layer.
        E.g. {0: [0.25, 0.75]} will split layer 0 into 2 layers, with first
        layer equal to 0.25 of original thickness and second layer 0.75 of
        original thickness.
    layer : str, optional
        name of layer dimension, by default 'layer'
    top : str, optional
        name of data variable containing top of layers, by default 'top'
    bot : str, optional
        name of data variable containing bottom of layers, by default 'bot'

    Returns
    -------
    new_top, new_bot : xarray.DataArrays
        DataArrays containing new tops and bottoms after splitting layers.
    reindexer : OrderedDict
        dictionary mapping new to old layer indices.
    """

    # calculate thickness
    thickness, top3d = calculate_thickness(ds, top=top, bot=bot)

    # calculate new number of layers
    new_nlay = ds[layer].size + \
        sum([len(sf) for sf in split_dict.values()]) - len(split_dict)

    # create new DataArrays for storing new top/bot
    new_bot = xr.DataArray(data=np.nan,
                           dims=["layer", "y", "x"],
                           coords={"layer": np.arange(new_nlay),
                                   "y": ds.y.data,
                                   "x": ds.x.data})
    new_top = xr.DataArray(data=np.nan,
                           dims=["layer", "y", "x"],
                           coords={"layer": np.arange(new_nlay),
                                   "y": ds.y.data,
                                   "x": ds.x.data})

    # dict to keep track of old and new layer indices
    reindexer = OrderedDict()

    j = 0  # new layer index
    isplit = 0  # split layer index

    # loop over original layers
    for i in range(ds[layer].size):

        # check if layer should be split
        if i in split_dict:

            # set new top based on old top
            if top3d:
                new_top.data[j] = ds[top].data[i]
            else:
                if i == 0:
                    new_top.data[j] = ds[top].data
                else:
                    new_top.data[j] = ds[bot].data[i - 1]

            # get split factors
            sf = split_dict[i]

            # check if factors add up to 1
            if np.sum(sf) != 1.0:
                raise ValueError(
                    "Sum of split factors for layer must equal 1.0!")
            logger.debug(f"{i}: Split layer {i} into {len(sf)} layers "
                         f"with fractions: {sf}")

            # loop over split factors
            for isf, factor in enumerate(sf):
                logger.debug(f"  - {isf}: Calculate new top/bot for "
                             f"new layer index {j}")

                # calculate new bot and new top
                new_bot.data[j] = new_top.data[j] - (factor * thickness[i])
                new_top.data[j + 1] = new_bot.data[j]

                # store new and old layer index
                reindexer[j] = i

                # increase new index
                j += 1

            # go to next layer to split
            isplit += 1

        # no split, remap old layer to new layer index
        else:
            logger.debug(f"{i:2d}: No split: map layer {i} to "
                         f"new layer index {j}")
            if top3d:
                new_top.data[j] = ds[top].data[i]
            else:
                if i == 0:
                    new_top.data[j] = ds[top].data.squeeze()
                else:
                    new_top.data[j] = ds[bot].data[i - 1]

            new_bot.data[j] = ds[bot].data[i]
            reindexer[j] = i
            j += 1

    return new_top, new_bot, reindexer


def fill_data_split_layers(da, reindexer):
    """Fill data for split layers with values from original layer.

    Parameters
    ----------
    da : xarray.DataArray or numpy.ndarray
        original array with data
    reindexer : dict
        dictionary containing mapping between new layer index and
        original layer index.

    Returns
    -------
    da_new : xarray.DataArray or numpy.ndarray
        array with filled data for split layers
    """
    if isinstance(da, xr.DataArray):
        da_new = xr.DataArray(data=np.nan,
                              dims=["layer", "y", "x"],
                              coords={"layer": np.arange(
                                  list(reindexer.keys())[-1] + 1),
                                  "y": da["y"],
                                  "x": da["x"]})
        for k, v in reindexer.items():
            da_new.data[k] = da.data[v]
    elif isinstance(da, np.ndarray):
        da_new = np.zeros((list(reindexer.keys())[-1] + 1), *da.shape[1:])
        for k, v in reindexer.items():
            da_new[k] = da[v]
    else:
        raise TypeError(f"Cannot fill type: '{type(da)}'!")
    return da_new


def split_layers_ds(ds, split_dict, layer='layer',
                    top='top', bot='bot', kh='kh', kv='kv'):
    """Split layers based in Dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        xarray Dataset containing information about layers
        (layers, top and bot)
    split_dict : dict
        dictionary with index of layers to split as keys and iterable
        of fractions that add up to 1 to indicate how to split up layer.
        E.g. {0: [0.25, 0.75]} will split layer 0 into 2 layers, with first
        layer equal to 0.25 of original thickness and second layer 0.75 of
        original thickness.
    layer : str, optional
        name of layer dimension, by default 'layer'
    top : str, optional
        name of data variable containing top of layers, by default 'top'
    bot : str, optional
        name of data variable containing bottom of layers, by default 'bot'
    kh : str, opti
        name of data variable containg horizontal hydraulic conductivity,
        by default 'kh'
    kv : str, optional
        name of data variable containg vertical hydraulic conductivity,
        by default 'kv'

    Returns
    -------
    ds_split : xarray.Dataset
        Dataset with new tops and bottoms taking into account split layers,
        and filled data for hydraulic conductivities.
    """

    parsed_dv = set([top, bot, kh, kv])

    dropped_dv = set(ds.data_vars.keys()) - parsed_dv
    if len(dropped_dv) > 0:
        logger.warning("Warning! Following data variables "
                       f"will be dropped: {dropped_dv}")

    # calculate new tops/bots
    logger.info("Calculating new layer tops and bottoms...")

    new_top, new_bot, reindexer = layer_split_top_bot(
        ds, split_dict, layer=layer, top=top, bot=bot)

    # fill kh/kv
    logger.info(f"Fill value '{kh}' for split layers with "
                "value original layer.")
    da_kh = fill_data_split_layers(ds["kh"], reindexer)
    logger.info(f"Fill value '{kv}' for split layers with "
                "value original layer.")
    da_kv = fill_data_split_layers(ds["kv"], reindexer)

    # get new layer names
    layer_names = []
    for j, i in reindexer.items():
        layercode = ds[layer].data[i]

        if layercode in layer_names:
            if isinstance(layercode, str):
                ilay = np.sum([1 for ilay in layer_names
                               if ilay.startswith(layercode)]) + 1
                layercode += f"_{ilay}"
            else:
                layercode = j

        layer_names.append(layercode)

    # assign new layer names
    new_top = new_top.assign_coords(layer=layer_names)
    new_bot = new_bot.assign_coords(layer=layer_names)
    da_kh = da_kh.assign_coords(layer=layer_names)
    da_kv = da_kv.assign_coords(layer=layer_names)

    # add reindexer to attributes
    attrs = ds.attrs.copy()
    attrs["split_reindexer"] = reindexer

    # create new dataset
    logger.info("Done! Created new dataset with split layers!")
    ds_split = xr.Dataset({"top": new_top,
                           "bot": new_bot,
                           "kh": da_kh,
                           "kv": da_kv},
                          attrs=attrs)

    return ds_split


def layer_combine_top_bot(ds, combine_layers, layer='layer',
                          top='top', bot='bot'):
    """Calculate new tops and bottoms for combined layers.

    Parameters
    ----------
    ds : xarray.Dataset
        xarray Dataset containing information about layers
        (layers, top and bot)
    combine_layers : list of tuple of ints
        list of tuples, with each tuple containing integers indicating
        layer indices to combine into one layer. E.g. [(0, 1), (2, 3)] will
        combine layers 0 and 1 into a single layer (with index 0) and layers
        2 and 3 into a second layer (with index 1).
    layer : str, optional
        name of layer dimension, by default 'layer'
    top : str, optional
        name of data variable containing top of layers, by default 'top'
    bot : str, optional
        name of data variable containing bottom of layers, by default 'bot'

    Returns
    -------
    new_top, new_bot : xarray.DataArrays
        DataArrays containing new tops and bottoms after splitting layers.
    reindexer : OrderedDict
        dictionary mapping new to old layer indices.
    """
    # calculate new number of layers
    new_nlay = (ds[layer].size -
                sum([len(c) for c in combine_layers]) +
                len(combine_layers))

    # create new DataArrays for storing new top/bot
    new_bot = xr.DataArray(data=np.nan,
                           dims=["layer", "y", "x"],
                           coords={"layer": np.arange(new_nlay),
                                   "y": ds.y.data,
                                   "x": ds.x.data})
    new_top = xr.DataArray(data=np.nan,
                           dims=["layer", "y", "x"],
                           coords={"layer": np.arange(new_nlay),
                                   "y": ds.y.data,
                                   "x": ds.x.data})

    # dict to keep track of old and new layer indices
    reindexer = OrderedDict()

    j = 0  # new layer index
    icomb = 0  # combine layer index

    # loop over original layers
    for i in range(ds.layer.size):

        # check whether to combine layers
        if i in np.concatenate(combine_layers):
            # get indices of layers
            c = combine_layers[icomb]
            # store new and original layer indices
            reindexer[j] = c
            # only need to calculate new top/bot once for each merged layer
            if i == np.min(c):
                logger.debug(f"{j:2d}: Merge layers {c} as layer {j}, "
                             "calculate new top/bot.")
                tops = ds[top].data[c, :, :]
                bots = ds[bot].data[c, :, :]
                new_top.data[j] = np.nanmax(tops, axis=0)
                new_bot.data[j] = np.nanmin(bots, axis=0)

            elif i == np.max(c):
                # advance combine layer index after merging layers
                icomb += 1
                # advance new layer index
                j += 1
                continue
            else:
                # no need to merge more than once, so continue loop
                continue
        else:
            # do not merge, only map old layer index to new layer index
            logger.debug(f"{j:2d}: Do not merge, map old layer index "
                         "to new layer index.")
            new_top.data[j] = ds[top].data[i]
            new_bot.data[j] = ds[bot].data[i]
            reindexer[j] = i
            j += 1

    return new_top, new_bot, reindexer


def sum_param_combined_layers(da, reindexer):
    """Calculate combined layer parameter with sum.

    Parameters
    ----------
    da : xarray.DataArray
        data array to calculate combined parameters for
    reindexer : OrderedDict
        dictionary mapping new layer indices to old layer indices

    Returns
    -------
    da_new : xarray.DataArray
        data array containing new parameters for combined layers and old
        parameters for unmodified layers.
    """
    da_new = xr.DataArray(data=np.nan,
                          dims=["layer", "y", "x"],
                          coords={"layer": np.arange(
                              list(reindexer.keys())[-1] + 1),
                              "y": da["y"],
                              "x": da["x"]})

    for k, v in reindexer.items():
        if isinstance(v, tuple):
            psum = np.sum(da.data[v, :, :], axis=0)
        else:
            psum = da.data[v]
        da_new.data[k] = psum
    return da_new


def kheq_combined_layers(kh, thickness, reindexer):
    """Calculate equivalent horizontal hydraulic conductivity.

    Parameters
    ----------
    kh : xarray.DataArray
        data array containing horizontal hydraulic conductivity
    thickness : xarray.DataArray
            data array containing thickness per layer
    reindexer : OrdererDict
        dictionary mapping new layer indices to old layer indices

    Returns
    -------
    da_kh : xarray.DataArray
        data array containing equivalent horizontal hydraulic conductivity
        for combined layers and original hydraulic conductivity in unmodified
        layers
    """
    da_kh = xr.DataArray(data=np.nan,
                         dims=["layer", "y", "x"],
                         coords={"layer": np.arange(
                             list(reindexer.keys())[-1] + 1),
                             "y": kh["y"],
                             "x": kh["x"]})

    for k, v in reindexer.items():
        if isinstance(v, tuple):
            kheq = np.nansum(thickness.data[v, :, :] * kh.data[v, :, :], axis=0) / \
                np.nansum(thickness.data[v, :, :], axis=0)
        else:
            kheq = kh.data[v]
        da_kh.data[k] = kheq
    return da_kh


def kveq_combined_layers(kv, thickness, reindexer):
    """Calculate equivalent vertical hydraulic conductivity.

    Parameters
    ----------
    kv : xarray.DataArray
        data array containing vertical hydraulic conductivity
    thickness : xarray.DataArray
        data array containing thickness per layer
    reindexer : OrdererDict
        dictionary mapping new layer indices to old layer indices

    Returns
    -------
    da_kv : xarray.DataArray
        data array containing equivalent vertical hydraulic conductivity
        for combined layers and original hydraulic conductivity in unmodified
        layers
    """
    da_kv = xr.DataArray(data=np.nan,
                         dims=["layer", "y", "x"],
                         coords={"layer": np.arange(
                             list(reindexer.keys())[-1] + 1),
                             "y": kv["y"],
                             "x": kv["x"]})

    for k, v in reindexer.items():
        if isinstance(v, tuple):
            kveq = np.nansum(thickness.data[v, :, :], axis=0) / \
                np.nansum(thickness.data[v, :, :] / kv.data[v, :, :], axis=0)
        else:
            kveq = kv.data[v]
        da_kv.data[k] = kveq
    return da_kv


def combine_layers_ds(ds, combine_layers, layer='layer',
                      top='top', bot='bot',
                      kh="kh", kv="kv", kD="kD", c="c"):
    """Combine layers in Dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        xarray Dataset containing information about layers
        (layers, top and bot)
    combine_layers : list of tuple of ints
        list of tuples, with each tuple containing integers indicating
        layer indices to combine into one layer. E.g. [(0, 1), (2, 3)] will
        combine layers 0 and 1 into a single layer (with index 0) and layers
        2 and 3 into a second layer (with index 1).
    layer : str, optional
        name of layer dimension, by default 'layer'
    top : str, optional
        name of data variable containing top of layers, by default 'top'
    bot : str, optional
        name of data variable containing bottom of layers, by default 'bot'
    kh : str, optional
        name of data variable containg horizontal hydraulic conductivity,
        by default 'kh'. Not parsed if set to None.
    kv : str, optional
        name of data variable containg vertical hydraulic conductivity,
        by default 'kv'. Not parsed if set to None.
    kD : str, optional
        name of data variable containg transmissivity or kD,
        by default 'kD'. Not parsed if set to None.
    c : str, optional
        name of data variable containg resistance or c,
        by default 'c'. Not parsed if set to None.

    Returns
    -------
    ds_combine : xarray.Dataset
        Dataset with new tops and bottoms taking into account combined layers,
        and recalculated values for parameters (kh, kv, kD, c).
    """

    data_vars = []
    for dv in [kh, kv, kD, c]:
        if dv is not None:
            data_vars.append(dv)
    parsed_dv = set([top, bot] + data_vars)

    dropped_dv = set(ds.data_vars.keys()) - parsed_dv
    if len(dropped_dv) > 0:
        logger.warning("Warning! Following data variables "
                       f"will be dropped: {dropped_dv}")

    # calculate new tops/bots
    logger.info("Calculating new layer tops and bottoms...")

    da_dict = {}

    new_top, new_bot, reindexer = layer_combine_top_bot(
        ds, combine_layers, layer=layer, top=top, bot=bot)
    da_dict[top] = new_top
    da_dict[bot] = new_bot

    # calculate original thickness
    thickness, _ = calculate_thickness(ds, top=top, bot=bot)

    # calculate equivalent kh/kv
    if kh is not None:
        logger.info(f"Calculate equivalent '{kh}' for combined layers.")
        da_dict[kh] = kheq_combined_layers(ds[kh], thickness, reindexer)
    if kv is not None:
        logger.info(f"Calculate equivalent '{kv}' for combined layers.")
        da_dict[kv] = kveq_combined_layers(ds[kv], thickness, reindexer)
    if kD is not None:
        logger.info(f"Calculate value '{kD}' for combined layers with sum.")
        da_dict[kD] = sum_param_combined_layers(ds[kD], reindexer)
    if c is not None:
        logger.info(f"Calculate value '{c}' for combined layers with sum.")
        da_dict[c] = sum_param_combined_layers(ds[c], reindexer)

    # get new layer names, based on first sub-layer from each combined layer
    layer_names = []
    for _, i in reindexer.items():
        if isinstance(i, tuple):
            i = i[0]
        layercode = ds[layer].data[i]
        layer_names.append(layercode)

    # assign new layer names
    for k, da in da_dict.items():
        da_dict[k] = da.assign_coords(layer=layer_names)

    # add reindexer to attributes
    attrs = ds.attrs.copy()
    attrs["combine_reindexer"] = reindexer

    # create new dataset
    logger.info("Done! Created new dataset with combined layers!")
    ds_combine = xr.Dataset(da_dict,
                            attrs=attrs)

    return ds_combine


def add_kh_kv_from_ml_layer_to_dataset(ml_layer_ds, model_ds, anisotropy,
                                       fill_value_kh, fill_value_kv):
    """add kh and kv from a model layer dataset to the model dataset.

    Supports structured and vertex grids.

    Parameters
    ----------
    ml_layer_ds : xarray.Dataset
        dataset with model layer data with kh and kv
    model_ds : xarray.Dataset
        dataset with model data where kh and kv are added to
    anisotropy : int or float
        factor to calculate kv from kh or the other way around
    fill_value_kh : int or float, optional
        use this value for kh if there is no data in regis. The default is 1.0.
    fill_value_kv : int or float, optional
        use this value for kv if there is no data in regis. The default is 1.0.

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data with new kh and kv

    Notes
    -----
    some model dataset, such as regis, also have 'c' and 'kd' values. These
    are ignored at the moment
    """
    model_ds.attrs['anisotropy'] = anisotropy
    model_ds.attrs['fill_value_kh'] = fill_value_kh
    model_ds.attrs['fill_value_kv'] = fill_value_kv
    kh_arr = ml_layer_ds['kh'].data
    kv_arr = ml_layer_ds['kv'].data

    logger.info('add kh and kv from model layer dataset to modflow model')

    kh, kv = get_kh_kv(kh_arr, kv_arr, anisotropy,
                       fill_value_kh=fill_value_kh,
                       fill_value_kv=fill_value_kv)

    if model_ds.gridtype == 'structured':
        da_ones = util.get_da_from_da_ds(model_ds, dims=('layer', 'y', 'x'),
                                         data=1)
    elif model_ds.gridtype == 'vertex':
        da_ones = util.get_da_from_da_ds(model_ds, dims=('layer', 'cid'),
                                         data=1)
    else:
        raise ValueError(
            'function only support structured or vertex gridtypes')

    model_ds['kh'] = da_ones * kh

    model_ds['kv'] = da_ones * kv

    # keep attributes for bot en top
    for datavar in ['kh', 'kv']:
        for key, att in ml_layer_ds[datavar].attrs.items():
            model_ds[datavar].attrs[key] = att

    return model_ds


def get_kh_kv(kh_in, kv_in, anisotropy,
              fill_value_kh=1.0, fill_value_kv=1.0):
    """maak kh en kv rasters voor flopy vanuit een regis raster met nan
    waardes.

    vul kh raster door:
    1. pak kh uit regis, tenzij nan dan:
    2. pak kv uit regis vermenigvuldig met anisotropy, tenzij nan dan:
    3. pak fill_value_kh

    vul kv raster door:
    1. pak kv uit regis, tenzij nan dan:
    2. pak kh uit regis deel door anisotropy, tenzij nan dan:
    3. pak fill_value_kv

    Supports structured and vertex grids.

    Parameters
    ----------
    kh_in : np.ndarray
        kh from regis with nan values shape(nlay, nrow, ncol) or
        shape(nlay, len(cid))
    kv_in : np.ndarray
        kv from regis with nan values shape(nlay, nrow, ncol) or
        shape(nlay, len(cid))
    anisotropy : int or float
        factor to calculate kv from kh or the other way around
    fill_value_kh : int or float, optional
        use this value for kh if there is no data in regis. The default is 1.0.
    fill_value_kv : int or float, optional
        use this value for kv if there is no data in regis. The default is 1.0.

    Returns
    -------
    kh_out : np.ndarray
        kh without nan values (nlay, nrow, ncol) or shape(nlay, len(cid))
    kv_out : np.ndarray
        kv without nan values (nlay, nrow, ncol) or shape(nlay, len(cid))
    """
    kh_out = np.zeros_like(kh_in)
    for i, kh_lay in enumerate(kh_in):
        kh_new = kh_lay.copy()
        kv_new = kv_in[i].copy()
        if ~np.all(np.isnan(kh_new)):
            logger.debug(f'layer {i} has a kh')
            kh_out[i] = np.where(np.isnan(kh_new), kv_new * anisotropy, kh_new)
            kh_out[i] = np.where(np.isnan(kh_out[i]), fill_value_kh, kh_out[i])
        elif ~np.all(np.isnan(kv_new)):
            logger.debug(f'layer {i} has a kv')
            kh_out[i] = np.where(
                np.isnan(kv_new), fill_value_kh, kv_new * anisotropy)
        else:
            logger.info(f'kv and kh both undefined in layer {i}')
            kh_out[i] = fill_value_kh

    kv_out = np.zeros_like(kv_in)
    for i, kv_lay in enumerate(kv_in):
        kv_new = kv_lay.copy()
        kh_new = kh_in[i].copy()
        if ~np.all(np.isnan(kv_new)):
            logger.debug(f'layer {i} has a kv')
            kv_out[i] = np.where(np.isnan(kv_new), kh_new / anisotropy, kv_new)
            kv_out[i] = np.where(np.isnan(kv_out[i]), fill_value_kv, kv_out[i])
        elif ~np.all(np.isnan(kh_new)):
            logger.debug(f'layer {i} has a kh')
            kv_out[i] = np.where(
                np.isnan(kh_new), fill_value_kv, kh_new / anisotropy)
        else:
            logger.info(f'kv and kh both undefined in layer {i}')
            kv_out[i] = fill_value_kv

    return kh_out, kv_out


def fill_top_bot_kh_kv_at_mask(model_ds, fill_mask,
                               gridtype='structured',
                               gridprops=None):
    """Fill values in top, bot, kh and kv.

    Fill where:
    1. the cell is True in fill_mask
    2. the cell thickness is greater than 0

    Fill values:
    - top: 0
    - bot: minimum of bottom_filled or top
    - kh: kh_filled if thickness is greater than 0
    - kv: kv_filled if thickness is greater than 0

    Parameters
    ----------
    model_ds : xr.DataSet
        model dataset, should contain 'first_active_layer'
    fill_mask : xr.DataArray
        1 where a cell should be replaced by masked value.
    gridtype : str, optional
        type of grid.
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen. Default is None

    Returns
    -------
    model_ds : xr.DataSet
        model dataset with adjusted data variables: 'top', 'bot', 'kh', 'kv'
    """

    # zee cellen hebben altijd een top gelijk aan 0
    model_ds['top'].values = np.where(fill_mask, 0, model_ds['top'])

    if gridtype == 'structured':
        fill_function = resample.fillnan_dataarray_structured_grid
        fill_function_kwargs = {}
    elif gridtype == 'vertex':
        fill_function = resample.fillnan_dataarray_vertex_grid
        fill_function_kwargs = {'gridprops': gridprops}

    for lay in range(model_ds.dims['layer']):
        bottom_nan = xr.where(fill_mask, np.nan, model_ds['bot'][lay])
        bottom_filled = fill_function(bottom_nan, **fill_function_kwargs)

        kh_nan = xr.where(fill_mask, np.nan, model_ds['kh'][lay])
        kh_filled = fill_function(kh_nan, **fill_function_kwargs)

        kv_nan = xr.where(fill_mask, np.nan, model_ds['kv'][lay])
        kv_filled = fill_function(kv_nan, **fill_function_kwargs)

        if lay == 0:
            # top ligt onder bottom_filled -> laagdikte wordt 0
            # top ligt boven bottom_filled -> laagdikte o.b.v. bottom_filled
            mask_top = model_ds['top'] < bottom_filled
            model_ds['bot'][lay] = xr.where(fill_mask * mask_top,
                                            model_ds['top'],
                                            bottom_filled)
            model_ds['kh'][lay] = xr.where(fill_mask * mask_top,
                                           model_ds['kh'][lay],
                                           kh_filled)
            model_ds['kv'][lay] = xr.where(fill_mask * mask_top,
                                           model_ds['kv'][lay],
                                           kv_filled)

        else:
            # top ligt onder bottom_filled -> laagdikte wordt 0
            # top ligt boven bottom_filled -> laagdikte o.b.v. bottom_filled
            mask_top = model_ds['bot'][lay - 1] < bottom_filled
            model_ds['bot'][lay] = xr.where(fill_mask * mask_top,
                                            model_ds['bot'][lay - 1],
                                            bottom_filled)
            model_ds['kh'][lay] = xr.where(fill_mask * mask_top,
                                           model_ds['kh'][lay],
                                           kh_filled)
            model_ds['kv'][lay] = xr.where(fill_mask * mask_top,
                                           model_ds['kv'][lay],
                                           kv_filled)

    return model_ds


def update_model_ds_from_ml_layer_ds(model_ds, ml_layer_ds,
                                     gridtype='structured',
                                     gridprops=None,
                                     keep_vars=None,
                                     add_northsea=True,
                                     anisotropy=10,
                                     fill_value_kh=1.,
                                     fill_value_kv=0.1,
                                     cachedir=None):
    """Update a model dataset with a model layer dataset.

    Steps:

    1. Add the data variables in 'keep_vars' from the model layer dataset
    to the model dataset
    2. add the attributes of the model layer dataset to the model dataset if
    they don't exist yet.
    3. compute idomain from the bot values in the model layer dataset, add
    to model dataset
    4. compute top and bots from model layer dataset, add to model dataset
    5. compute kh, kv from model layer dataset, add to model dataset
    6. if gridtype is vertex add top, bot and area to gridprops
    7. if add_northsea is True:
        a) get cells from modelgrid that are within the northsea, add data
        variable 'northsea' to model_ds
        b) fill top, bot, kh and kv add northsea cell by extrapolation
        c) get bathymetry (northsea depth) from jarkus. Add datavariable
        bathymetry to model dataset


    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data, preferably without a grid definition.
    ml_layer_ds : xarray.Dataset
        dataset with model layer data corresponding to the modelgrid
    gridtype : str, optional
        type of grid, default is 'structured'
    gridprops : dictionary
        dictionary with grid properties output from gridgen.
    keep_vars : list of str
        variables in ml_layer_ds that will be used in model_ds
    add_northsea : bool, optional
        if True the nan values at the northsea are filled using the
        bathymetry from jarkus
    anisotropy : int or float
        factor to calculate kv from kh or the other way around
    fill_value_kh : int or float, optional
        use this value for kh if there is no data in regis. The default is 1.0.
    fill_value_kv : int or float, optional
        use this value for kv if there is no data in regis. The default is 1.0.
    cachedir : str, optional
        directory to store cached values, if None a temporary directory is
        used. default is None

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data
    """
    model_ds.attrs['gridtype'] = gridtype

    if keep_vars is None:
        keep_vars = []
    else:
        # update variables
        model_ds.update(ml_layer_ds[keep_vars])
        # update attributes
        for key, item in ml_layer_ds.attrs.items():
            if key not in model_ds.attrs.keys():
                model_ds.attrs.update({key: item})

    model_ds = add_idomain_from_bottom_to_dataset(ml_layer_ds['bot'],
                                                  model_ds)

    model_ds = add_top_bot_to_model_ds(ml_layer_ds, model_ds,
                                       gridtype=gridtype)

    model_ds = add_kh_kv_from_ml_layer_to_dataset(ml_layer_ds,
                                                  model_ds,
                                                  anisotropy,
                                                  fill_value_kh,
                                                  fill_value_kv)

    if gridtype == 'vertex':
        gridprops['top'] = model_ds['top'].data
        gridprops['botm'] = model_ds['bot'].data

        # add surface area of each cell
        model_ds['area'] = ('cid', gridprops.pop('area')
                            [:len(model_ds['cid'])])
    else:
        gridprops = None

    if add_northsea:
        logger.info(
            'nan values at the northsea are filled using the bathymetry from jarkus')

        # find grid cells with northsea
        model_ds.update(rws.get_northsea(model_ds,
                                         gridprops=gridprops,
                                         cachedir=cachedir,
                                         cachename='sea_model_ds.nc'))

        # fill top, bot, kh, kv at sea cells
        fill_mask = (model_ds['first_active_layer'] ==
                     model_ds.nodata) * model_ds['northsea']
        model_ds = fill_top_bot_kh_kv_at_mask(model_ds, fill_mask,
                                              gridtype=gridtype,
                                              gridprops=gridprops)

        # add bathymetry noordzee
        model_ds.update(jarkus.get_bathymetry(model_ds,
                                              model_ds['northsea'],
                                              gridprops=gridprops,
                                              cachedir=cachedir,
                                              cachename='bathymetry_model_ds.nc'))

        model_ds = jarkus.add_bathymetry_to_top_bot_kh_kv(model_ds,
                                                          model_ds['bathymetry'],
                                                          fill_mask)

        # update idomain on adjusted tops and bots
        model_ds['thickness'], _ = calculate_thickness(model_ds)
        model_ds['idomain'] = update_idomain_from_thickness(model_ds['idomain'],
                                                            model_ds['thickness'],
                                                            model_ds['northsea'])
        model_ds['first_active_layer'] = mgrid.get_first_active_layer_from_idomain(
            model_ds['idomain'])
        if gridtype == 'vertex':
            gridprops['top'] = model_ds['top'].data
            gridprops['botm'] = model_ds['bot'].data

    else:
        model_ds['thickness'], _ = calculate_thickness(model_ds)
        model_ds['first_active_layer'] = mgrid.get_first_active_layer_from_idomain(
            model_ds['idomain'])

    return model_ds


def add_idomain_from_bottom_to_dataset(bottom, model_ds, nodata=-999):
    """add idomain and first_active_layer to model_ds The active layers are
    defined as the layers where the bottom is not nan.

    Parameters
    ----------
    bottom : xarray.DataArray
        DataArray with bottom values of each layer. Nan values indicate
        inactive cells.
    model_ds : xarray.Dataset
        dataset with model data where idomain and first_active_layer
        are added to.
    nodata : int, optional
        nodata value used in integer arrays. For float arrays np.nan is use as
        nodata value. The default is -999.

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data including idomain and first_active_layer
    """
    logger.info('get active cells (idomain) from bottom DataArray')

    idomain = xr.where(bottom.isnull(), -1, 1)

    # if the top cell is inactive set idomain = 0, for other inactive cells
    # set idomain = -1
    idomain[0] = xr.where(idomain[0] == -1, 0, idomain[0])
    for i in range(1, bottom.shape[0]):
        idomain[i] = xr.where((idomain[i - 1] == 0)
                              & (idomain[i] == -1), 0, idomain[i])

    model_ds['idomain'] = idomain
    model_ds['first_active_layer'] = mgrid.get_first_active_layer_from_idomain(idomain,
                                                                               nodata=nodata)

    model_ds.attrs['nodata'] = nodata

    return model_ds


def update_idomain_from_thickness(idomain, thickness, mask):
    """get new idomain from thickness in the cells where mask is 1 (or True).

    Idomain becomes:
    -    1: if cell thickness is bigger than 0
    -    0: if cell thickness is 0 and it is the top layer
    -   -1: if cell thickness is 0 and the layer is in between active cells

    Parameters
    ----------
    idomain : xr.DataArray
        raster with idomain of each cell. dimensions should be (layer, y,x) or
        (layer, cid).
    thickness : xr.DataArray
        raster with thickness of each cell. dimensions should be (layer, y,x) or
        (layer, cid).
    mask : xr.DataArray
        raster with ones in cell where the ibound is adjusted. dimensions
        should be (y,x) or (cid).

    Returns
    -------
    idomain : xr.DataArray
        raster with adjusted idomain of each cell. dimensions should be
        (layer, y,x) or (layer, cid).
    """

    for ilay, thick in enumerate(thickness):
        if ilay == 0:
            mask1 = (thick == 0) * mask
            idomain[ilay] = xr.where(mask1, 0, idomain[ilay])
            mask2 = (thick > 0) * mask
            idomain[ilay] = xr.where(mask2, 1, idomain[ilay])
        else:
            mask1 = (thick == 0) * mask * (idomain[ilay - 1] == 0)
            idomain[ilay] = xr.where(mask1, 0, idomain[ilay])

            mask2 = (thick == 0) * mask * (idomain[ilay - 1] != 0)
            idomain[ilay] = xr.where(mask2, -1, idomain[ilay])

            mask3 = (thick != 0) * mask
            idomain[ilay] = xr.where(mask3, 1, idomain[ilay])

    return idomain


def add_top_bot_to_model_ds(ml_layer_ds, model_ds,
                            nodata=None,
                            gridtype='structured'):
    """add top and bot from a model layer dataset to THE model dataset.

    Supports structured and vertex grids.

    Parameters
    ----------
    ml_layer_ds : xarray.Dataset
        dataset with model layer data with a top and bottom
    model_ds : xarray.Dataset
        dataset with model data where top and bot are added to
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers. If nodata is None the
        nodata value in model_ds is used.
        the default is None
    gridtype : str, optional
        type of grid, options are 'structured' and 'vertex'.
        The default is 'structured'.

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data including top and bottom
    """
    if nodata is None:
        nodata = model_ds.attrs['nodata']

    logger.info(
        'using top and bottom from model layers dataset for modflow model')
    logger.info('replace nan values for inactive layers with dummy value')

    if gridtype == 'structured':
        model_ds = add_top_bot_structured(ml_layer_ds, model_ds,
                                          nodata=nodata)

    elif gridtype == 'vertex':
        model_ds = add_top_bot_vertex(ml_layer_ds, model_ds,
                                      nodata=nodata)

    return model_ds


def add_top_bot_vertex(ml_layer_ds, model_ds, nodata=-999):
    """Voeg top en bottom vanuit layer dataset toe aan de model dataset.

    Deze functie is bedoeld voor vertex arrays in modflow 6. Supports
    only vertex grids.

    Stappen:

    1. Zorg dat de onderste laag altijd een bodemhoogte heeft, als de bodem
       van alle bovenliggende lagen nan is, pak dan 0.
    2. Zorg dat de top van de bovenste laag altijd een waarde heeft, als de
       top van alle onderligende lagen nan is, pak dan 0.
    3. Vul de nan waarden in alle andere lagen door:
        a) pak bodem uit regis, tenzij nan dan:
        b) gebruik bodem van de laag erboven (of de top voor de bovenste laag)

    Parameters
    ----------
    ml_layer_ds : xarray.Dataset
        dataset with model layer data with a top and bottom
    model_ds : xarray.Dataset
        dataset with model data where top and bottom are added to
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data including top and bottom
    """
    # step 1:
    # set nan-value in bottom array
    # set to zero if value is nan in all layers
    # set to minimum value of all layers if there is any value in any layer
    active_domain = model_ds['first_active_layer'].data != nodata

    lowest_bottom = ml_layer_ds['bot'].data[-1].copy()
    if np.any(~active_domain):
        percentage = 100 * (~active_domain).sum() / \
            (active_domain.shape[0])
        if percentage > 80:
            logger.warning(f'{percentage:0.1f}% of all cells have nan '
                           'values in every layer there is probably a '
                           'problem with your extent.')

        # set bottom to zero if bottom in a cell is nan in all layers
        lowest_bottom = np.where(active_domain, lowest_bottom, 0)

    if np.any(np.isnan(lowest_bottom)):
        # set bottom in a cell to lowest bottom of all layers
        i_nan = np.where(np.isnan(lowest_bottom))
        for i in i_nan:
            val = np.nanmin(ml_layer_ds['bot'].data[:, i])
            lowest_bottom[i] = val
            if np.isnan(val):
                raise ValueError(
                    'this should never happen please contact Artesia')

    # step 2: get highest top values of all layers without nan values
    highest_top = ml_layer_ds['top'].data[0].copy()
    if np.any(np.isnan(highest_top)):
        highest_top = np.where(active_domain, highest_top, 0)

    if np.any(np.isnan(highest_top)):
        i_nan = np.where(np.isnan(highest_top))
        for i in i_nan:
            val = np.nanmax(ml_layer_ds['top'].data[:, i])
            highest_top[i] = val
            if np.isnan(val):
                raise ValueError(
                    'this should never happen please contact Artesia')

    # step 3: fill nans in all layers
    nlay = model_ds.dims['layer']
    top_bot_raw = np.ones((nlay + 1, model_ds.dims['cid']))
    top_bot_raw[0] = highest_top
    top_bot_raw[1:-1] = ml_layer_ds['bot'].data[:-1].copy()
    top_bot_raw[-1] = lowest_bottom
    top_bot = np.ones_like(top_bot_raw)
    for i_from_bot, blay in enumerate(top_bot_raw[::-1]):
        i_from_top = nlay - i_from_bot
        new_lay = blay.copy()
        if np.any(np.isnan(new_lay)):
            lay_from_bot = i_from_bot
            lay_from_top = nlay - lay_from_bot
            while np.any(np.isnan(new_lay)):
                new_lay = np.where(np.isnan(new_lay),
                                   top_bot_raw[lay_from_top],
                                   new_lay)
                lay_from_bot += 1
                lay_from_top = nlay - lay_from_bot

        top_bot[i_from_top] = new_lay

    model_ds['bot'] = xr.DataArray(top_bot[1:], dims=('layer', 'cid'),
                                   coords={'cid': model_ds.cid.data,
                                           'layer': model_ds.layer.data})
    model_ds['top'] = xr.DataArray(top_bot[0], dims=('cid',),
                                   coords={'cid': model_ds.cid.data})

    # keep attributes for bot en top
    for datavar in ['top', 'bot']:
        for key, att in ml_layer_ds[datavar].attrs.items():
            model_ds[datavar].attrs[key] = att

    return model_ds


def add_top_bot_structured(ml_layer_ds, model_ds, nodata=-999):
    """Voeg top en bottom vanuit een layer dataset toe aan de model dataset.

    Deze functie is bedoeld voor structured arrays in modflow 6. Supports
    only structured grids.

    Stappen:

    1. Zorg dat de onderste laag altijd een bodemhoogte heeft, als de bodem
       van alle bovenliggende lagen nan is, pak dan 0.
    2. Zorg dat de top van de bovenste laag altijd een waarde heeft, als de
       top van alle onderligende lagen nan is, pak dan 0.
    3. Vul de nan waarden in alle andere lagen door:
        a) pak bodem uit de model layer dataset, tenzij nan dan:
        b) gebruik bodem van de laag erboven (of de top voor de bovenste laag)

    Parameters
    ----------
    ml_layer_ds : xarray.Dataset
        dataset with model layer data with a top and bottom
    model_ds : xarray.Dataset
        dataset with model data where top and bottom are added to
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data including top and bottom
    """

    active_domain = model_ds['first_active_layer'].data != nodata

    # step 1:
    # set nan-value in bottom array
    # set to zero if value is nan in all layers
    # set to minimum value of all layers if there is any value in any layer
    lowest_bottom = ml_layer_ds['bot'].data[-1].copy()
    if np.any(~active_domain):
        percentage = 100 * (~active_domain).sum() / \
            (active_domain.shape[0] * active_domain.shape[1])
        if percentage > 80:
            logger.warning(f'{percentage:0.1f}% of all cells have nan '
                           'values in every layer there is probably a '
                           'problem with your extent.')
        # set bottom to zero if bottom in a cell is nan in all layers
        lowest_bottom = np.where(active_domain, lowest_bottom, 0)

    if np.any(np.isnan(lowest_bottom)):
        # set bottom in a cell to lowest bottom of all layers
        rc_nan = np.where(np.isnan(lowest_bottom))
        for row, col in zip(rc_nan[0], rc_nan[1]):
            val = np.nanmin(ml_layer_ds['bot'].data[:, row, col])
            lowest_bottom[row, col] = val
            if np.isnan(val):
                raise ValueError(
                    'this should never happen please contact Onno')

    # step 2: get highest top values of all layers without nan values
    highest_top = ml_layer_ds['top'].data[0].copy()
    if np.any(np.isnan(highest_top)):
        # set top to zero if top in a cell is nan in all layers
        highest_top = np.where(active_domain, highest_top, 0)

    if np.any(np.isnan(highest_top)):
        # set top in a cell to highest top of all layers
        rc_nan = np.where(np.isnan(highest_top))
        for row, col in zip(rc_nan[0], rc_nan[1]):
            val = np.nanmax(ml_layer_ds['top'].data[:, row, col])
            highest_top[row, col] = val
            if np.isnan(val):
                raise ValueError(
                    'this should never happen please contact Onno')

    # step 3: fill nans in all layers
    nlay = model_ds.dims['layer']
    nrow = model_ds.dims['y']
    ncol = model_ds.dims['x']
    top_bot_raw = np.ones((nlay + 1, nrow, ncol))
    top_bot_raw[0] = highest_top
    top_bot_raw[1:-1] = ml_layer_ds['bot'].data[:-1].copy()
    top_bot_raw[-1] = lowest_bottom
    top_bot = np.ones_like(top_bot_raw)
    for i_from_bot, blay in enumerate(top_bot_raw[::-1]):
        i_from_top = nlay - i_from_bot
        new_lay = blay.copy()
        if np.any(np.isnan(new_lay)):
            lay_from_bot = i_from_bot
            lay_from_top = nlay - lay_from_bot
            while np.any(np.isnan(new_lay)):
                new_lay = np.where(np.isnan(new_lay),
                                   top_bot_raw[lay_from_top],
                                   new_lay)
                lay_from_bot += 1
                lay_from_top = nlay - lay_from_bot

        top_bot[i_from_top] = new_lay

    model_ds['bot'] = xr.DataArray(top_bot[1:], dims=('layer', 'y', 'x'),
                                   coords={'x': model_ds.x.data,
                                           'y': model_ds.y.data,
                                           'layer': model_ds.layer.data})

    model_ds['top'] = xr.DataArray(top_bot[0], dims=('y', 'x'),
                                   coords={'x': model_ds.x.data,
                                           'y': model_ds.y.data})

    # keep attributes for bot en top
    for datavar in ['top', 'bot']:
        for key, att in ml_layer_ds[datavar].attrs.items():
            model_ds[datavar].attrs[key] = att

    return model_ds



# -*- coding: utf-8 -*-
"""Module containing model grid functions.

-   project data on different grid types
-   obtain various types of rec_lists from a grid that
    can be used as input for a MODFLOW package
-   fill, interpolate and resample grid data
"""
import copy
import logging
import os
import sys

import flopy
import pandas as pd
import geopandas as gpd
import numpy as np
import shapely
import xarray as xr
from flopy.discretization.structuredgrid import StructuredGrid
from flopy.utils.gridgen import Gridgen
from flopy.utils.gridintersect import GridIntersect
from shapely.prepared import prep
from tqdm import tqdm
from scipy.interpolate import griddata

from shapely.geometry import Point
from .. import cache, mfpackages, util

logger = logging.getLogger(__name__)


def modelgrid_from_model_ds(model_ds, gridprops=None):
    """Get flopy modelgrid from model_ds.

    Parameters
    ----------
    model_ds : xarray DataSet
        model dataset.
    gridprops : dict, optional
        extra model properties when using vertex grids.
        The default is None.

    Returns
    -------
    modelgrid : StructuredGrid, VertexGrid
        grid information.
    """

    if model_ds.gridtype == 'structured':
        if not isinstance(model_ds.extent, (tuple, list, np.ndarray)):
            raise TypeError(
                f'extent should be a list, tuple or numpy array, not {type(model_ds.extent)}')

        modelgrid = StructuredGrid(delc=np.array([model_ds.delc] * model_ds.dims['y']),
                                   delr=np.array([model_ds.delr] *
                                                 model_ds.dims['x']),
                                   xoff=model_ds.extent[0], yoff=model_ds.extent[2])
    elif model_ds.gridtype == 'vertex':
        _, gwf = mfpackages.sim_tdis_gwf_ims_from_model_ds(model_ds)
        # somehow this function modifies gridprops['vertices'] from a list of
        # lists into a list of tuples, for type checking later on I don't
        # want this, therefore I make a deepcopy here
        flopy.mf6.ModflowGwfdisv(gwf, idomain=model_ds['idomain'].data,
                                 **copy.deepcopy(gridprops))
        modelgrid = gwf.modelgrid

    return modelgrid


def get_xy_mid_structured(extent, delr, delc, descending_y=True):
    """Calculates the x and y coordinates of the cell centers of a structured
    grid.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax)
    delr : int or float,
        cell size along rows, equal to dx
    delc : int or float,
        cell size along columns, equal to dy
    descending_y : bool, optional
        if True the resulting ymid array is in descending order. This is the
        default for MODFLOW models. default is True.

    Returns
    -------
    xmid : np.array
        x-coordinates of the cell centers shape(ncol)
    ymid : np.array
        y-coordinates of the cell centers shape(nrow)
    """
    # check if extent is valid
    if (extent[1] - extent[0]) % delr != 0.0:
        raise ValueError('invalid extent, the extent should contain an integer'
                         ' number of cells in the x-direction')
    if (extent[3] - extent[2]) % delc != 0.0:
        raise ValueError('invalid extent, the extent should contain an integer'
                         ' number of cells in the y-direction')

    # get cell mids
    x_mid_start = extent[0] + 0.5 * delr
    x_mid_end = extent[1] - 0.5 * delr
    y_mid_start = extent[2] + 0.5 * delc
    y_mid_end = extent[3] - 0.5 * delc

    ncol = int((extent[1] - extent[0]) / delr)
    nrow = int((extent[3] - extent[2]) / delc)

    xmid = np.linspace(x_mid_start, x_mid_end, ncol)
    if descending_y:
        ymid = np.linspace(y_mid_end, y_mid_start, nrow)
    else:
        ymid = np.linspace(y_mid_start, y_mid_end, nrow)

    return xmid, ymid


@cache.cache_pklz
def create_vertex_grid(model_name, gridgen_ws, gwf=None,
                       refine_features=None, extent=None,
                       nlay=None, nrow=None, ncol=None,
                       delr=None, delc=None, exe_name=None):
    """Create vertex grid. Refine grid using refinement features.

    Parameters
    ----------
    gridgen_ws : str
        directory to save gridgen files.
    model_name : str
        name of the model.
    gwf : flopy.mf6.ModflowGwf
        groundwater flow model, if structured grid is already defined
        parameters defining the grid are taken from modelgrid if not
        explicitly passed.
    refine_features : list of tuples, optional
        list of tuples containing refinement features, tuples must each
        contain [(geometry, shape_type, level)]. Geometry can be a path
        pointing to a shapefile or an object defining the geometry.
        For accepted types for each entry, see
        `flopy.utils.gridgen.Gridgen.add_refinement_features()`
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax) of the desired grid.
    nlay : int, optional
        number of model layers. If not passed,
    nrow : int, optional
        number of model rows.
    ncol : int, optional
        number of model columns
    delr : int or float, optional
        cell size along rows of the desired grid (dx).
    delc : int or float, optional
        cell size along columns of the desired grid (dy).
    exe_name : str
        Filepath to the gridgen executable

    Returns
    -------
    gridprops : dictionary
        gridprops with the vertex grid information.
    """

    logger.info('create vertex grid using gridgen')

    # if existing structured grid, take parameters from grid if not
    # explicitly passed
    if gwf is not None:
        if gwf.modelgrid.grid_type == "structured":
            nlay = gwf.modelgrid.nlay if nlay is None else nlay
            nrow = gwf.modelgrid.nrow if nrow is None else nrow
            ncol = gwf.modelgrid.ncol if ncol is None else ncol
            delr = gwf.modelgrid.delr if delr is None else delr
            delc = gwf.modelgrid.delc if delc is None else delc
            extent = gwf.modelgrid.extent if extent is None else extent

    # create temporary groundwaterflow model with dis package
    if gwf is not None:
        _gwf_temp = copy.deepcopy(gwf)
    else:
        _sim_temp = flopy.mf6.MFSimulation()
        _gwf_temp = flopy.mf6.MFModel(_sim_temp)
    _dis_temp = flopy.mf6.ModflowGwfdis(_gwf_temp, pname='dis',
                                        nlay=nlay, nrow=nrow, ncol=ncol,
                                        delr=delr, delc=delc,
                                        xorigin=extent[0],
                                        yorigin=extent[2],
                                        filename='{}.dis'.format(model_name))

    # Define new default `exe_name` for NHFLO
    if exe_name is None:
        exe_name = os.path.join(os.path.dirname(__file__),
                                '..', '..', 'bin', 'gridgen')

        if sys.platform.startswith('win'):
            exe_name += ".exe"

    g = Gridgen(_dis_temp, model_ws=gridgen_ws, exe_name=exe_name)

    if refine_features is not None:
        for shp_fname, shp_type, lvl in refine_features:
            if isinstance(shp_fname, str):
                shp_fname = os.path.relpath(shp_fname, gridgen_ws)
                if shp_fname.endswith('.shp'):
                    shp_fname = shp_fname[:-4]
            g.add_refinement_features(shp_fname, shp_type, lvl, range(nlay))

    g.build()

    gridprops = g.get_gridprops_disv()
    gridprops['area'] = g.get_area()

    return gridprops


def get_xyi_cid(gridprops=None, model_ds=None):
    """Get x and y coördinates of the cell mids from the cellids in the grid
    properties.

    Parameters
    ----------
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen. If gridprops is
        None xyi and cid will be obtained from model_ds.
    model_ds : xarray.Dataset
        dataset with model data. Should have dimension (layer, cid).

    Returns
    -------
    xyi : numpy.ndarray
        array with x and y coördinates of cell centers, shape(len(cid), 2).
    cid : numpy.ndarray
        array with cellids, shape(len(cid))
    """
    if gridprops is not None:
        xc_gwf = [cell2d[1] for cell2d in gridprops['cell2d']]
        yc_gwf = [cell2d[2] for cell2d in gridprops['cell2d']]
        xyi = np.vstack((xc_gwf, yc_gwf)).T
        cid = np.array([c[0] for c in gridprops['cell2d']])
    elif model_ds is not None:
        xyi = np.array(list(zip(model_ds.x.values, model_ds.y.values)))
        cid = model_ds.cid.values
    else:
        raise ValueError('either gridprops or model_ds should be specified')

    return xyi, cid


def col_to_list(col_in, model_ds, cellids):
    """Convert array data in model_ds to a list of values for specific cells.

    This function is typically used to create a rec_array with stress period
    data for the modflow packages. Can be used for structured and
    vertex grids.

    Parameters
    ----------
    col_in : str, int or float
        if col_in is a str type it is the name of the column in model_ds.
        if col_in is an int or a float it is a value that will be used for all
        cells in cellids.
    model_ds : xarray.Dataset
        dataset with model data. Can have dimension (layer, y, x) or
        (layer, cid).
    cellids : tuple of numpy arrays
        tuple with indices of the cells that will be used to create the list
        with values. There are 3 options:
            1.   cellids contains (layers, rows, columns)
            2.   cellids contains (rows, columns) or (layers, cids)
            3.   cellids contains (cids)

    Raises
    ------
    ValueError
        raised if the cellids are in the wrong format.

    Returns
    -------
    col_lst : list
        raster values from model_ds presented in a list per cell.
    """

    if isinstance(col_in, str):
        if len(cellids) == 3:
            # 3d grid
            col_lst = [model_ds[col_in].data[lay, row, col]
                       for lay, row, col in zip(cellids[0], cellids[1], cellids[2])]
        elif len(cellids) == 2:
            # 2d grid or vertex 3d grid
            col_lst = [model_ds[col_in].data[row, col]
                       for row, col in zip(cellids[0], cellids[1])]
        elif len(cellids) == 1:
            # 2d vertex grid
            col_lst = model_ds[col_in].data[cellids[0]]
        else:
            raise ValueError(
                f'could not create a column list for col_in={col_in}')
    else:
        col_lst = [col_in] * len(cellids[0])

    return col_lst


def lrc_to_rec_list(layers, rows, columns, cellids, model_ds,
                    col1=None, col2=None, col3=None):
    """Create a rec list for stress period data from a set of cellids.

    Used for structured grids.


    Parameters
    ----------
    layers : list or numpy.ndarray
        list with the layer for each cell in the rec_list.
    rows : list or numpy.ndarray
        list with the rows for each cell in the rec_list.
    columns : list or numpy.ndarray
        list with the columns for each cell in the rec_list.
    cellids : tuple of numpy arrays
        tuple with indices of the cells that will be used to create the list
        with values.
    model_ds : xarray.Dataset
        dataset with model data. Can have dimension (layer, y, x) or
        (layer, cid).
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column)) for each row.

        col1 should be the following value for each package (can also be the
            name of a timeseries):
            rch: recharge [L/T]
            ghb: head [L]
            drn: drain level [L]
            chd: head [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1) for each row.

        col2 should be the following value for each package (can also be the
            name of a timeseries):
            ghb: conductance [L^2/T]
            drn: conductance [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1, col2) for each row.

        col3 should be the following value for each package (can also be the
            name of a timeseries):

    Raises
    ------
    ValueError
        Question: will this error ever occur?.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer,row,column), col1, col2, col3).
    """

    if col1 is None:
        rec_list = list(zip(zip(layers, rows, columns)))
    elif (col1 is not None) and col2 is None:
        col1_lst = col_to_list(col1, model_ds, cellids)
        rec_list = list(zip(zip(layers, rows, columns),
                            col1_lst))
    elif (col2 is not None) and col3 is None:
        col1_lst = col_to_list(col1, model_ds, cellids)
        col2_lst = col_to_list(col2, model_ds, cellids)
        rec_list = list(zip(zip(layers, rows, columns),
                            col1_lst, col2_lst))
    elif (col3 is not None):
        col1_lst = col_to_list(col1, model_ds, cellids)
        col2_lst = col_to_list(col2, model_ds, cellids)
        col3_lst = col_to_list(col3, model_ds, cellids)
        rec_list = list(zip(zip(layers, rows, columns),
                            col1_lst, col2_lst, col3_lst))
    else:
        raise ValueError(
            'invalid combination of values for col1, col2 and col3')

    return rec_list


def data_array_3d_to_rec_list(model_ds, mask,
                              col1=None, col2=None, col3=None,
                              only_active_cells=True):
    """Create a rec list for stress period data from a model dataset.

    Used for structured grids.


    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data and dimensions (layer, y, x)
    mask : xarray.DataArray for booleans
        True for the cells that will be used in the rec list.
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column)) for each row.

        col1 should be the following value for each package (can also be the
            name of a timeseries):
            rch: recharge [L/T]
            ghb: head [L]
            drn: drain level [L]
            chd: head [L]
            riv: stage [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1) for each row.

        col2 should be the following value for each package (can also be the
            name of a timeseries):
            ghb: conductance [L^2/T]
            drn: conductance [L^2/T]
            riv: conductance [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1, col2) for each row.

        col3 should be the following value for each package (can also be the
            name of a timeseries):
            riv: river bottom [L]

    only_active_cells : bool, optional
        If True an extra mask is used to only include cells with an idomain
        of 1. The default is True.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer,row,column), col1, col2, col3).
    """
    if only_active_cells:
        cellids = np.where((mask) & (model_ds['idomain'] == 1))
    else:
        cellids = np.where(mask)

    layers = cellids[0]
    rows = cellids[1]
    columns = cellids[2]

    rec_list = lrc_to_rec_list(layers, rows, columns, cellids, model_ds,
                               col1, col2, col3)

    return rec_list


def data_array_2d_to_rec_list(model_ds, mask,
                              col1=None, col2=None, col3=None,
                              layer=0,
                              first_active_layer=False,
                              only_active_cells=True):
    """Create a rec list for stress period data from a model dataset.

    Used for structured grids.


    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data and dimensions (layer, y, x)
    mask : xarray.DataArray for booleans
        True for the cells that will be used in the rec list.
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column)) for each row.

        col1 should be the following value for each package (can also be the
            name of a timeseries):
            rch: recharge [L/T]
            ghb: head [L]
            drn: drain level [L]
            chd: head [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1) for each row.

        col2 should be the following value for each package (can also be the
            name of a timeseries):
            ghb: conductance [L^2/T]
            drn: conductance [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        ((layer,row,column), col1, col2) for each row.

        col3 should be the following value for each package (can also be the
            name of a timeseries):
    layer : int, optional
        layer used in the rec_list. Not used if first_active_layer is True.
        default is 0
    first_active_layer : bool, optional
        If True an extra mask is applied to use the first active layer of each
        cell in the grid. The default is False.
    only_active_cells : bool, optional
        If True an extra mask is used to only include cells with an idomain
        of 1. The default is True.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer,row,column), col1, col2, col3).
    """

    if first_active_layer:
        if 'first_active_layer' not in model_ds:
            model_ds['first_active_layer'] = get_first_active_layer_from_idomain(
                model_ds['idomain'])

        cellids = np.where(
            (mask) & (model_ds['first_active_layer'] != model_ds.nodata))
        layers = col_to_list('first_active_layer', model_ds, cellids)
    elif only_active_cells:
        cellids = np.where((mask) & (model_ds['idomain'][layer] == 1))
        layers = col_to_list(layer, model_ds, cellids)
    else:
        cellids = np.where(mask)
        layers = col_to_list(layer, model_ds, cellids)

    rows = cellids[-2]
    columns = cellids[-1]

    rec_list = lrc_to_rec_list(layers, rows, columns, cellids, model_ds,
                               col1, col2, col3)

    return rec_list


def lcid_to_rec_list(layers, cellids, model_ds,
                     col1=None, col2=None, col3=None):
    """Create a rec list for stress period data from a set of cellids.

    Used for vertex grids.


    Parameters
    ----------
    layers : list or numpy.ndarray
        list with the layer for each cell in the rec_list.
    cellids : tuple of numpy arrays
        tuple with indices of the cells that will be used to create the list
        with values for a column. There are 2 options:
            1. cellids contains (layers, cids)
            2. cellids contains (cids)
    model_ds : xarray.Dataset
        dataset with model data. Should have dimensions (layer, cid).
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,cid)) for each row. col1 should be the following value for
        each package (can also be the name of a timeseries):
        -   rch: recharge [L/T]
        -   ghb: head [L]
        -   drn: drain level [L]
        -   chd: head [L]
        -   riv: stage [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        ((layer,cid), col1) for each row. col2 should be the following
        value for each package (can also be the name of a timeseries):
        -   ghb: conductance [L^2/T]
        -   drn: conductance [L^2/T]
        -   riv: conductacnt [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        ((layer,cid), col1, col2) for each row. col3 should be the following
        value for each package (can also be the name of a timeseries):
        -   riv: bottom [L]

    Raises
    ------
    ValueError
        Question: will this error ever occur?.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer, cid), col1, col2, col3)
        grids.
    """
    if col1 is None:
        rec_list = list(zip(zip(layers, cellids[-1])))
    elif (col1 is not None) and col2 is None:
        col1_lst = col_to_list(col1, model_ds, cellids)
        rec_list = list(zip(zip(layers, cellids[-1]),
                            col1_lst))
    elif (col2 is not None) and col3 is None:
        col1_lst = col_to_list(col1, model_ds, cellids)
        col2_lst = col_to_list(col2, model_ds, cellids)
        rec_list = list(zip(zip(layers, cellids[-1]),
                            col1_lst, col2_lst))
    elif (col3 is not None):
        col1_lst = col_to_list(col1, model_ds, cellids)
        col2_lst = col_to_list(col2, model_ds, cellids)
        col3_lst = col_to_list(col3, model_ds, cellids)
        rec_list = list(zip(zip(layers, cellids[-1]),
                            col1_lst, col2_lst, col3_lst))
    else:
        raise ValueError(
            'invalid combination of values for col1, col2 and col3')

    return rec_list


def data_array_2d_vertex_to_rec_list(model_ds, mask,
                                     col1=None, col2=None, col3=None,
                                     only_active_cells=True):
    """Create a rec list for stress period data from a model dataset.

    Used for vertex grids.


    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data and dimensions (layer, cid)
    mask : xarray.DataArray for booleans
        True for the cells that will be used in the rec list.
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,cid)) for each row.

        col1 should be the following value for each package (can also be the
            name of a timeseries):
            rch: recharge [L/T]
            ghb: head [L]
            drn: drain level [L]
            chd: head [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        (((layer,cid), col1) for each row.

        col2 should be the following value for each package (can also be the
            name of a timeseries):
            ghb: conductance [L^2/T]
            drn: conductance [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        (((layer,cid), col1, col2) for each row.

        col3 should be the following value for each package (can also be the
            name of a timeseries):
            riv: bottom [L]
    only_active_cells : bool, optional
        If True an extra mask is used to only include cells with an idomain
        of 1. The default is True.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer,row,column), col1, col2, col3).
    """
    if only_active_cells:
        cellids = np.where((mask) & (model_ds['idomain'] == 1))
    else:
        cellids = np.where(mask)

    layers = cellids[0]

    rec_list = lcid_to_rec_list(layers, cellids, model_ds,
                                col1, col2, col3)

    return rec_list


def data_array_1d_vertex_to_rec_list(model_ds, mask,
                                     col1=None, col2=None, col3=None,
                                     layer=0,
                                     first_active_layer=False,
                                     only_active_cells=True):
    """Create a rec list for stress period data from a model dataset.

    Used for vertex grids.


    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data and dimensions (layer, cid)
    mask : xarray.DataArray for booleans
        True for the cells that will be used in the rec list.
    col1 : str, int or float, optional
        1st column of the rec_list, if None the rec_list will be a list with
        ((layer,cid)) for each row.

        col1 should be the following value for each package (can also be the
            name of a timeseries):
            rch: recharge [L/T]
            ghb: head [L]
            drn: drain level [L]
            chd: head [L]

    col2 : str, int or float, optional
        2nd column of the rec_list, if None the rec_list will be a list with
        (((layer,cid), col1) for each row.

        col2 should be the following value for each package (can also be the
            name of a timeseries):
            ghb: conductance [L^2/T]
            drn: conductance [L^2/T]

    col3 : str, int or float, optional
        3th column of the rec_list, if None the rec_list will be a list with
        (((layer,cid), col1, col2) for each row.

        col3 should be the following value for each package (can also be the
            name of a timeseries):
            riv: bottom [L]
    layer : int, optional
        layer used in the rec_list. Not used if first_active_layer is True.
        default is 0
    first_active_layer : bool, optional
        If True an extra mask is applied to use the first active layer of each
        cell in the grid. The default is False.
    only_active_cells : bool, optional
        If True an extra mask is used to only include cells with an idomain
        of 1. The default is True.

    Returns
    -------
    rec_list : list of tuples
        every row consist of ((layer,cid), col1, col2, col3).
    """
    if first_active_layer:
        cellids = np.where(
            (mask) & (model_ds['first_active_layer'] != model_ds.nodata))
        layers = col_to_list('first_active_layer', model_ds, cellids)
    elif only_active_cells:
        cellids = np.where((mask) & (model_ds['idomain'][layer] == 1))
        layers = col_to_list(layer, model_ds, cellids)
    else:
        cellids = np.where(mask)
        layers = col_to_list(layer, model_ds, cellids)

    rec_list = lcid_to_rec_list(layers, cellids, model_ds, col1, col2, col3)

    return rec_list


def polygon_to_area(modelgrid, polygon, da,
                    gridtype='structured'):
    """create a grid with the surface area in each cell based on a polygon
    value.

    Parameters
    ----------
    modelgrid : flopy.discretization.structuredgrid.StructuredGrid
        grid.
    polygon : shapely.geometry.polygon.Polygon
        polygon feature.
    da : xarray.DataArray
        data array that is filled with polygon data

    Returns
    -------
    area_array : xarray.DataArray
        area of polygon within each modelgrid cell
    """
    if polygon.type == 'Polygon':
        pass
    elif polygon.type == 'MultiPolygon':
        Warning(
            'function not tested for MultiPolygon type, can have unexpected results')
    else:
        raise TypeError(
            f'input geometry should by of type "Polygon" not {polygon.type}')

    ix = GridIntersect(modelgrid, method='vertex')
    opp_cells = ix.intersect(polygon)

    if gridtype == 'structured':
        area_array = util.get_da_from_da_ds(da, dims=('y', 'x'), data=0)
        for opp_row in opp_cells:
            area = opp_row[-2]
            area_array[opp_row[0][0], opp_row[0][1]] = area
    elif gridtype == 'vertex':
        area_array = util.get_da_from_da_ds(da, dims=('cid',), data=0)
        cids = opp_cells.cellids
        area = opp_cells.areas
        area_array[cids.astype(int)] = area

    return area_array


def gdf2data_array_struc(gdf, gwf,
                         field='VALUE',
                         agg_method=None,
                         interp_method=None):
    """ Project vector data on a structured grid. Aggregate data if multiple
    geometries are in a single cell

    Parameters
    ----------
    gdf : geopandas.GeoDataframe
        vector data can only contain a single geometry type.
    gwf : flopy groundwater flow model
        model with a structured grid.
    field : str, optional
        column name in the geodataframe. The default is 'VALUE'.
    interp_method : str or None, optional
        method to obtain values in cells without geometry by interpolating
        between cells with values. Options are 'nearest' and 'linear'.
    agg_method : str, optional
        aggregation method to handle multiple geometries in one cell, options
        are:
        - max, min, mean,
        - length_weighted (lines), max_length (lines),
        - area_weighted (polygon), area_max (polygon).
        The default is 'max'.

    Returns
    -------
    da : xarray DataArray
        DESCRIPTION.

    """
    xmid = gwf.modelgrid.get_xcellcenters_for_layer(0)[0]
    ymid = gwf.modelgrid.get_ycellcenters_for_layer(0)[:, 0]
    da = xr.DataArray(np.nan, dims=('y', 'x'),
                      coords={'y': ymid, 'x': xmid})

    # interpolate data
    if interp_method is not None:
        arr = interpolate_gdf_to_array(gdf, gwf, field=field,
                                       method=interp_method)
        da.values = arr

        return da

    gdf_cellid = gdf2grid(gdf, gwf, 'vertex')

    if gdf_cellid.cellid.duplicated().any():
        # aggregate data
        if agg_method is None:
            raise ValueError('multiple geometries in one cell please define aggregation method')
        gdf_agg = aggregate_vector_per_cell(gdf_cellid, {field: agg_method},
                                            gwf)
    else:
        # aggregation not neccesary
        gdf_agg = gdf_cellid[[field]]
        gdf_agg.set_index(pd.MultiIndex.from_tuples(gdf_cellid.cellid.values),
                          inplace=True)

    for ind, row in gdf_agg.iterrows():
        da.values[ind[0], ind[1]] = row[field]

    return da


def interpolate_gdf_to_array(gdf, gwf, field='values', method='nearest'):
    """ interpolate data from a point gdf


    Parameters
    ----------
    gdf : geopandas.GeoDataframe
        vector data can only contain a single geometry type.
    gwf : flopy groundwater flow model
        model with a structured grid.
    field : str, optional
        column name in the geodataframe. The default is 'values'.
    method : str or None, optional
        method to obtain values in cells without geometry by interpolating
        between cells with values. Options are 'nearest' and 'linear'.

    Returns
    -------
    arr : np.array
        numpy array with interpolated data.

    """
    # check geometry
    geom_types = gdf.geometry.type.unique()
    if geom_types[0] != 'Point':
        raise NotImplementedError('can only use interpolation with point geometries')

    # check field
    if field not in gdf.columns:
        raise ValueError(f"Missing column in DataFrame: {field}")

    points = np.array([[g.x, g.y] for g in gdf.geometry])
    values = gdf[field].values
    xi = np.vstack((gwf.modelgrid.xcellcenters.flatten(),
                    gwf.modelgrid.ycellcenters.flatten())).T
    vals = griddata(points, values, xi, method=method)
    arr = np.reshape(vals, (gwf.modelgrid.nrow, gwf.modelgrid.ncol))

    return arr


def _agg_max_area(gdf, col):
    return gdf.loc[gdf.area.idxmax(), col]


def _agg_area_weighted(gdf, col):
    nanmask = gdf[col].isna()
    aw = ((gdf.area * gdf[col]).sum(skipna=True)
           / gdf.loc[~nanmask].area.sum())
    return aw


def _agg_max_length(gdf, col):
    return gdf.loc[gdf.length.idxmax(), col]


def _agg_length_weighted(gdf, col):
    nanmask = gdf[col].isna()
    aw = ((gdf.length * gdf[col]).sum(skipna=True)
           / gdf.loc[~nanmask].length.sum())
    return aw


def _agg_nearest(gdf, col, gwf):
    cid = gdf['cellid'].values[0]
    cellcenter = Point(gwf.modelgrid.xcellcenters[0][cid[1]],
                       gwf.modelgrid.ycellcenters[:, 0][cid[0]])
    val = gdf.iloc[gdf.distance(cellcenter).argmin()].loc[col]
    return val


def _get_aggregates_values(group, fields_methods, gwf=None):

    agg_dic = {}
    for field, method in fields_methods.items():
        # aggregation is only necesary if group shape is greater than 1
        if group.shape[0] == 1:
            agg_dic[field] = group[field].values[0]
        if method == 'max':
            agg_dic[field] = group[field].max()
        elif method == 'min':
            agg_dic[field] = group[field].min()
        elif method == 'mean':
            agg_dic[field] = group[field].mean()
        elif method == 'nearest':
            agg_dic[field] = _agg_nearest(group, field, gwf)
        elif method == 'length_weighted':  # only for lines
            agg_dic[field] = _agg_length_weighted(group, field)
        elif method == 'max_length':  # only for lines
            agg_dic[field] = _agg_max_length(group, field)
        elif method == "area_weighted":  # only for polygons
            agg_dic[field] = _agg_area_weighted(group, field)
        elif method == "max_area":  # only for polygons
            agg_dic[field] = _agg_max_area(group, field)
        elif method == 'center_grid':  # only for polygons
            raise NotImplementedError
        else:
            raise ValueError(f"Method '{method}' not recognized!")

    return agg_dic


def aggregate_vector_per_cell(gdf, fields_methods, gwf=None):
    """Aggregate vector features per cell.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame containing points, lines or polygons per grid cell.
    fields_methods: dict
        fields (keys) in the Geodataframe with their aggregation method (items)
        aggregation methods can be:
        max, min, mean, length_weighted (lines), max_length (lines),
        area_weighted (polygon), area_max (polygon).
    gwf : flopy Groundwater flow model
        only necesary if one of the field methods is 'nearest'

    Returns
    -------
    celldata : pd.DataFrame
        DataFrame with aggregated surface water parameters per grid cell
    """
    # check geometry types
    geom_types = gdf.geometry.type.unique()
    if len(geom_types) > 1:
        if len(geom_types) == 2 and ('Polygon' in geom_types) and ('MultiPolygon' in geom_types):
            pass
        else:
            raise TypeError('cannot aggregate geometries of different types')
    if bool({'length_weighted', 'max_length'} & set(fields_methods.values())):
        assert geom_types[0] == 'LineString', 'can only use length methods with line geometries'
    if bool({'area_weighted', 'max_area'} & set(fields_methods.values())):
        if ('Polygon' in geom_types) or ('MultiPolygon' in geom_types):
            pass
        else:
            raise TypeError('can only use area methods with polygon geometries')

    # check fields
    missing_cols = set(fields_methods.keys()).difference(gdf.columns)
    if len(missing_cols) > 0:
        raise ValueError(f"Missing columns in DataFrame: {missing_cols}")

    # aggregate data
    gr = gdf.groupby(by="cellid")
    celldata = pd.DataFrame(index=gr.groups.keys())
    for cid, group in tqdm(gr, desc="Aggregate vector data"):
        agg_dic = _get_aggregates_values(group, fields_methods,
                                         gwf)
        for key, item in agg_dic.items():
            celldata.loc[cid, key] = item

    return celldata


def gdf_to_bool_data_array(gdf, mfgrid, model_ds):
    """convert a GeoDataFrame with polygon geometries into a data array
    corresponding to the modelgrid in which each cell is 1 (True) if one or
    more geometries are (partly) in that cell.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame or shapely.geometry
        shapes that will be rasterised.
    mfgrid : flopy grid
        model grid.
    model_ds : xr.DataSet
        xarray with model data

    Returns
    -------
    da : xr.DataArray
        1 if polygon is in cell, 0 otherwise. Grid dimensions according to
        model_ds and mfgrid.
    """

    # build list of gridcells
    ix = GridIntersect(mfgrid, method="vertex")

    if model_ds.gridtype == 'structured':
        da = util.get_da_from_da_ds(model_ds, dims=('y', 'x'), data=0)
    elif model_ds.gridtype == 'vertex':
        da = util.get_da_from_da_ds(model_ds, dims=('cid',), data=0)
    else:
        raise ValueError(
            'function only support structured or vertex gridtypes')

    if isinstance(gdf, gpd.GeoDataFrame):
        geoms = gdf.geometry.values
    elif isinstance(gdf, shapely.geometry.base.BaseGeometry):
        geoms = [gdf]

    for geom in geoms:
        # prepare shape for efficient batch intersection check
        prepshp = prep(geom)

        # get only gridcells that intersect
        filtered = filter(prepshp.intersects, ix._get_gridshapes())

        # cell ids for intersecting cells
        cids = [c.name for c in filtered]

        if model_ds.gridtype == 'structured':
            for cid in cids:
                da[cid[0], cid[1]] = 1
        elif model_ds.gridtype == 'vertex':
            da[cids] = 1

    return da


def gdf_to_bool_dataset(model_ds, gdf, mfgrid, da_name):
    """convert a GeoDataFrame with polygon geometries into a model dataset with
    a data_array named 'da_name' in which each cell is 1 (True) if one or more
    geometries are (partly) in that cell.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        polygon shapes with surface water.
    mfgrid : flopy grid
        model grid.
    model_ds : xr.DataSet
        xarray with model data

    Returns
    -------
    model_ds_out : xr.Dataset
        Dataset with a single DataArray, this DataArray is 1 if polygon is in
        cell, 0 otherwise. Grid dimensions according to model_ds and mfgrid.
    """
    model_ds_out = util.get_model_ds_empty(model_ds)
    model_ds_out[da_name] = gdf_to_bool_data_array(gdf, mfgrid, model_ds)

    return model_ds_out


def gdf2grid(gdf, ml, method=None, ix=None,
             desc="Intersecting with grid", **kwargs):
    """Cut a geodataframe gdf by the grid of a flopy modflow model ml. This
    method is just a wrapper around the GridIntersect method from flopy.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        A GeoDataFrame that needs to be cut by the grid. The GeoDataFrame can
        consist of multiple types (Point, LineString, Polygon and the Multi-
        variants).
    ml : flopy.modflow.Modflow or flopy.mf6.ModflowGwf
        The flopy model that defines the grid.
    method : string, optional
        Method passed to the GridIntersect-class. The default is None, which
        makes GridIntersect choose the best method.
    ix : flopy.utils.GridIntersect, optional
        GridIntersect, if not provided the modelgrid in ml is used.
    **kwargs : keyword arguments
        keyword arguments are passed to the intersect_*-methods.

    Returns
    -------
    geopandas.GeoDataFrame
        The GeoDataFrame with the geometries per grid-cell.
    """
    if ix is None:
        ix = flopy.utils.GridIntersect(ml.modelgrid, method=method)
    shps = []
    for _, shp in tqdm(gdf.iterrows(), total=gdf.shape[0], desc=desc):
        if hasattr(ix, 'intersect'):
            r = ix.intersect(shp.geometry, **kwargs)
        else:
            if shp.geometry.type in ['Point', 'MultiPoint']:
                r = ix.intersect_point(shp.geometry, **kwargs)
            elif shp.geometry.type in ['LineString', 'MultiLineString']:
                r = ix.intersect_linestring(shp.geometry, **kwargs)
            elif shp.geometry.type in ['Polygon', 'MultiPolygon']:
                r = ix.intersect_polygon(shp.geometry, **kwargs)
            else:
                msg = 'Unknown geometry type: {}'.format(shp.geometry.type)
                raise(Exception(msg))
        for i in range(r.shape[0]):
            shpn = shp.copy()
            shpn['cellid'] = r['cellids'][i]
            shpn.geometry = r['ixshapes'][i]
            shps.append(shpn)
    return gpd.GeoDataFrame(shps)


def get_thickness_from_topbot(top, bot):
    """get thickness from data arrays with top and bots.

    Parameters
    ----------
    top : xr.DataArray
        raster with top of each cell. dimensions should be (y,x) or (cid).
    bot : xr.DataArray
        raster with bottom of each cell. dimensions should be (layer, y,x) or
        (layer, cid).

    Returns
    -------
    thickness : xr.DataArray
        raster with thickness of each cell. dimensions should be (layer, y,x)
        or (layer, cid).
    """
    DeprecationWarning(
        'function is deprecated please use calculate_thickness function instead')

    if np.ndim(top) > 2:
        raise NotImplementedError('function works only for 2d top')

    # get thickness
    if bot.ndim == 3:
        thickness = util.get_da_from_da_ds(bot, dims=('layer', 'y', 'x'))
    elif bot.ndim == 2:
        thickness = util.get_da_from_da_ds(bot, dims=('layer', 'cid'))
    else:
        raise ValueError(
            'function only support structured or vertex gridtypes')

    for lay in range(len(bot)):
        if lay == 0:
            thickness[lay] = top - bot[lay]
        else:
            thickness[lay] = bot[lay - 1] - bot[lay]

    return thickness


def get_vertices(model_ds, modelgrid=None,
                 gridprops=None, vert_per_cid=4):
    """get vertices of a vertex modelgrid from the modelgrid or from the
    gridprops. Only return the 4 corners of each cell and not the corners of
    adjacent cells thus limiting the vertices per cell to 4 points.

    If the modelgrid is given the xvertices and yvertices attributes of the
    modelgrid are used. If the gridprops are given the cell2d and vertices are
    obtained from the gridprops.

    Parameters
    ----------
    model_ds : xr.DataSet
        model dataset, attribute grid_type should be 'vertex'
    modelgrid : flopy.discretization.vertexgrid.VertexGrid
        vertex grid with attributes xvertices and yvertices.
    gridprops : dictionary
        gridproperties obtained from gridgen
    vert_per_cid : int or None:
        number of vertices per cell:
        - 4 return the 4 vertices of each cell
        - 5 return the 4 vertices of each cell + one duplicate vertex
        (sometimes useful if you want to create polygons)
        - anything else, the maximum number of vertices. For locally refined
        cells this includes all the vertices adjacent to the cell.

        if vert_per_cid is 4 or 5 vertices are removed using the
        Ramer-Douglas-Peucker Algorithm -> https://github.com/fhirschmann/rdp.

    Returns
    -------
    vertices_da : xarray DataArray
         Vertex coördinates per cell with dimensions(cid, no_vert, 2).
    """

    # obtain
    if modelgrid is not None:
        xvert = modelgrid.xvertices
        yvert = modelgrid.yvertices
        if vert_per_cid == 4:
            from rdp import rdp
            vertices_arr = np.array([rdp(list(zip(xvert[i], yvert[i])))[
                                    :-1] for i in range(len(xvert))])
        elif vert_per_cid == 5:
            from rdp import rdp
            vertices_arr = np.array(
                [rdp(list(zip(xvert[i], yvert[i]))) for i in range(len(xvert))])
        else:
            raise NotImplementedError()

    elif gridprops is not None:
        all_vertices = np.ones(
            (len(gridprops['cell2d']), len(gridprops['cell2d'][0]), 2)) * np.nan
        for i, cell in enumerate(gridprops['cell2d']):
            for j in range(cell[3]):
                all_vertices[i, j] = gridprops['vertices'][cell[4 + j]][1:]

        if vert_per_cid in [4, 5]:
            from rdp import rdp
            clean_vertices = np.ones((len(gridprops['cell2d']), 5, 2)) * np.nan
            for i, vert in enumerate(all_vertices):
                clean_vertices[i] = rdp(vert[~np.isnan(vert).any(axis=1)])
            if vert_per_cid == 4:
                vertices_arr = clean_vertices[:, :4, :]
            else:
                vertices_arr = clean_vertices

        else:
            vertices_arr = all_vertices

    else:
        raise ValueError('Specify gridprops or modelgrid')

    vertices_da = xr.DataArray(vertices_arr,
                               dims=('cid', 'vert_per_cid', 'xy'),
                               coords={'cid': model_ds.cid.values,
                                       'vert_per_cid': range(vertices_arr.shape[1]),
                                       'xy': ['x', 'y']})

    return vertices_da


def get_first_active_layer_from_idomain(idomain, nodata=-999):
    """get the first active layer in each cell from the idomain.

    Parameters
    ----------
    idomain : xr.DataArray
        idomain. Shape can be (layer, y, x) or (layer, cid)
    nodata : int, optional
        nodata value. used for cells that are inactive in all layers.
        The default is -999.

    Returns
    -------
    first_active_layer : xr.DataArray
        raster in which each cell has the zero based number of the first
        active layer. Shape can be (y, x) or (cid)
    """
    logger.info('get first active modellayer for each cell in idomain')

    first_active_layer = xr.where(idomain[0] == 1, 0, nodata)
    for i in range(1, idomain.shape[0]):
        first_active_layer = xr.where((first_active_layer == nodata) & (idomain[i] == 1),
                                      i,
                                      first_active_layer)

    return first_active_layer




import datetime as dt
import logging
import os

import nlmod
import numpy as np
import pandas as pd
import xarray as xr

from .. import cache, mdims
from . import regis

logger = logging.getLogger(__name__)


def get_default_lithoklasse_translation_table():
    return pd.read_csv(os.path.join(nlmod.NLMOD_DATADIR,
                                    'geotop',
                                    'litho_eenheden.csv'),
                       index_col=0)


@cache.cache_netcdf
def get_geotop(extent, delr, delc,
               regis_ds, regis_layer='HLc'):
    """get a model layer dataset for modflow from geotop within a certain
    extent and grid.

    if regis_ds and regis_layer are defined the geotop model is only created
    to replace this regis_layer in a regis layer model.


    Parameters
    ----------
    extent : list, tuple or np.array
        desired model extent (xmin, xmax, ymin, ymax)
    delr : int or float,
        cell size along rows, equal to dx
    delc : int or float,
        cell size along columns, equal to dy
    regis_ds: xarray.DataSet
        regis dataset used to cut geotop to the same x and y coordinates
    regis_layer: str, optional
        layer of regis dataset that will be filled with geotop. The default is
        'HLc'.

    Returns
    -------
    geotop_ds: xr.DataSet
        geotop dataset with top, bot, kh and kv per geo_eenheid
    """
    # check extent
    extent2, _, _ = regis.fit_extent_to_regis(extent, delr, delc)
    for coord1, coord2 in zip(extent, extent2):
        if coord1 != coord2:
            raise ValueError(
                'extent not fitted to regis please fit to regis first, '
                'use the nlmod.regis.fit_extent_to_regis function')

    geotop_url = r'http://www.dinodata.nl/opendap/GeoTOP/geotop.nc'
    geotop_ds_raw1 = get_geotop_raw_within_extent(extent, geotop_url)

    litho_translate_df = pd.read_csv(os.path.join(nlmod.NLMOD_DATADIR,
                                                  'geotop',
                                                  'litho_eenheden.csv'),
                                     index_col=0)

    geo_eenheid_translate_df = pd.read_csv(os.path.join(nlmod.NLMOD_DATADIR,
                                                        'geotop',
                                                        'geo_eenheden.csv'),
                                           index_col=0,
                                           keep_default_na=False)

    geotop_ds_raw = convert_geotop_to_ml_layers(geotop_ds_raw1,
                                                regis_ds=regis_ds,
                                                regis_layer=regis_layer,
                                                litho_translate_df=litho_translate_df,
                                                geo_eenheid_translate_df=geo_eenheid_translate_df)

    logger.info('resample geotop data to structured modelgrid')
    geotop_ds = mdims.resample_dataset_to_structured_grid(geotop_ds_raw,
                                                          extent,
                                                          delr, delc)
    geotop_ds.attrs['extent'] = extent
    geotop_ds.attrs['delr'] = delr
    geotop_ds.attrs['delc'] = delc
    geotop_ds.attrs['gridtype'] = 'structured'

    for datavar in geotop_ds:
        geotop_ds[datavar].attrs['source'] = 'Geotop'
        geotop_ds[datavar].attrs['url'] = geotop_url
        geotop_ds[datavar].attrs['date'] = dt.datetime.now().strftime('%Y%m%d')
        if datavar in ['top', 'bot']:
            geotop_ds[datavar].attrs['units'] = 'mNAP'
        elif datavar in ['kh', 'kv']:
            geotop_ds[datavar].attrs['units'] = 'm/day'

    return geotop_ds


def get_geotop_raw_within_extent(extent, url):
    """Get a slice of the geotop netcdf url within the extent, set the x and y
    coordinates to match the cell centers and keep only the strat and lithok
    data variables.

    Parameters
    ----------
    extent : list, tuple or np.array
        desired model extent (xmin, xmax, ymin, ymax)
    url : str
        url of geotop netcdf file

    Returns
    -------
    geotop_ds_raw : xarray Dataset
        slices geotop netcdf.
    """

    geotop_ds_raw = xr.open_dataset(url)

    # set x and y dimensions to cell center
    for dim in ['x', 'y']:
        old_dim = geotop_ds_raw[dim].values
        geotop_ds_raw[dim] = old_dim + (old_dim[1] - old_dim[0]) / 2

    # slice extent
    geotop_ds_raw = geotop_ds_raw.sel(x=slice(extent[0], extent[1]),
                                      y=slice(extent[2], extent[3]))
    geotop_ds_raw = geotop_ds_raw[['strat', 'lithok']]

    return geotop_ds_raw


def convert_geotop_to_ml_layers(geotop_ds_raw1, regis_ds=None, regis_layer=None,
                                litho_translate_df=None,
                                geo_eenheid_translate_df=None):
    """does the following steps to obtain model layers based on geotop:

        1. slice by regis layer (if not None)
        2. compute kh from lithoklasse
        3. create a layer model based on geo-eenheden

    Parameters
    ----------
    geotop_ds_raw1: xr.Dataset
        dataset with geotop netcdf
    regis_ds: xarray.DataSet
        regis dataset used to cut geotop to the same x and y coördinates
    regis_layer: str, optional
        layer of regis dataset that will be filled with geotop
    litho_translate_df: pandas.DataFrame
        horizontal conductance (kh)
    geo_eenheid_translate_df: pandas.DataFrame
        dictionary to translate geo_eenheid to a geo name

    Returns
    -------
    geotop_ds_raw: xarray.DataSet
        geotop dataset with added horizontal conductance
    """

    # stap 1
    if (regis_ds is not None) and (regis_layer is not None):
        logger.info(f'slice geotop with regis layer {regis_layer}')
        top_rl = regis_ds['top'].sel(layer=regis_layer)
        bot_rl = regis_ds['bot'].sel(layer=regis_layer)

        geotop_ds_raw = geotop_ds_raw1.sel(z=slice(np.floor(bot_rl.min().data),
                                                   np.ceil(top_rl.max().data)))

    # stap 2 maak kh matrix a.d.v. lithoklasse
    logger.info('create kh matrix from lithoklasse and csv file')
    kh_from_litho = xr.zeros_like(geotop_ds_raw.lithok)
    for i, row in litho_translate_df.iterrows():
        kh_from_litho = xr.where(geotop_ds_raw.lithok == i,
                                 row['hor_conductance_default'],
                                 kh_from_litho)
    geotop_ds_raw['kh_from_litho'] = kh_from_litho

    # stap 3 maak een laag per geo-eenheid
    geotop_ds_mod = get_top_bot_from_geo_eenheid(geotop_ds_raw,
                                                 geo_eenheid_translate_df)

    return geotop_ds_mod


def get_top_bot_from_geo_eenheid(geotop_ds_raw, geo_eenheid_translate_df):
    """get top, bottom and kh of each geo-eenheid in geotop dataset.

    Parameters
    ----------
    geotop_ds_raw: xr.DataSet
        geotop dataset with added horizontal conductance
    geo_eenheid_translate_df: pandas.DataFrame
        dictionary to translate geo_eenheid to a geo name

    Returns
    -------
    geotop_ds_mod: xr.DataSet
        geotop dataset with top, bot, kh and kv per geo_eenheid

    Note
    ----
    the 'geo_eenheid' >6000 are 'stroombanen' these are difficult to add because
    they can occur above and below any other 'geo_eenheid' therefore they are
    added to the geo_eenheid below the stroombaan.
    """

    # vindt alle geo-eenheden in model_extent
    geo_eenheden = np.unique(geotop_ds_raw.strat.data)
    geo_eenheden = geo_eenheden[np.isfinite(geo_eenheden)]
    stroombaan_eenheden = geo_eenheden[geo_eenheden < 5999]
    geo_eenheden = geo_eenheden[geo_eenheden < 5999]

    # geo eenheid 2000 zit boven 1130
    if (2000. in geo_eenheden) and (1130. in geo_eenheden):
        geo_eenheden[(geo_eenheden == 2000.) +
                     (geo_eenheden == 1130.)] = [2000., 1130.]

    geo_names = [geo_eenheid_translate_df.loc[float(
        geo_eenh), 'Code (lagenmodel en boringen)'] for geo_eenh in geo_eenheden]

    # fill top and bot
    top = np.ones(
        (geotop_ds_raw.y.shape[0], geotop_ds_raw.x.shape[0], len(geo_names))) * np.nan
    bot = np.ones(
        (geotop_ds_raw.y.shape[0], geotop_ds_raw.x.shape[0], len(geo_names))) * np.nan
    lay = 0
    logger.info('creating top and bot per geo eenheid')
    for geo_eenheid in geo_eenheden:
        logger.info(geo_eenheid)

        mask = geotop_ds_raw.strat == geo_eenheid
        geo_z = xr.where(mask, geotop_ds_raw.z, np.nan)

        top[:, :, lay] = geo_z.max(dim='z').T + 0.5
        bot[:, :, lay] = geo_z.min(dim='z').T

        lay += 1

    geotop_ds_mod = add_stroombanen_and_get_kh(geotop_ds_raw, top, bot,
                                               geo_names)

    geotop_ds_mod.attrs['stroombanen'] = stroombaan_eenheden

    return geotop_ds_mod


def add_stroombanen_and_get_kh(geotop_ds_raw, top, bot, geo_names,
                               f_anisotropy=0.25):
    """add stroombanen to tops and bots of geo_eenheden, also computes kh per
    geo_eenheid. Kh is computed by taking the average of all kh's of a
    geo_eenheid within a cell (e.g. if one geo_eenheid has a thickness of 1,5m
    in a certain cell the kh of the cell is calculated as the mean of the 3
    cells in geotop)

    Parameters
    ----------
    geotop_ds_raw: xr.DataSet
        geotop dataset with added horizontal conductance
    top: np.array
        raster with top of each geo_eenheid, shape(nlay,nrow,ncol)
    bot: np.array
        raster with bottom of each geo_eenheid, shape(nlay,nrow,ncol)
    geo_names: list of str
        names of each geo_eenheid
    f_anisotropy: float, optional
        anisotropy factor kv/kh, ratio between vertical and horizontal
        hydraulic conductivities, by default 0.25.


    Returns
    -------
    geotop_ds_mod: xr.DataSet
        geotop dataset with top, bot, kh and kv per geo_eenheid
    """
    kh = np.ones(
        (geotop_ds_raw.y.shape[0], geotop_ds_raw.x.shape[0], len(geo_names))) * np.nan
    thickness = np.ones(
        (geotop_ds_raw.y.shape[0], geotop_ds_raw.x.shape[0], len(geo_names))) * np.nan
    z = xr.ones_like(geotop_ds_raw.lithok) * geotop_ds_raw.z
    logger.info('adding stroombanen to top and bot of each layer')
    logger.info('get kh for each layer')

    for lay in range(top.shape[2]):
        logger.info(geo_names[lay])
        if lay == 0:
            top[:, :, 0] = np.nanmax(top, axis=2)
        else:
            top[:, :, lay] = bot[:, :, lay - 1]
        bot[:, :, lay] = np.where(
            np.isnan(bot[:, :, lay]), top[:, :, lay], bot[:, :, lay])
        thickness[:, :, lay] = (top[:, :, lay] - bot[:, :, lay])

        # check which geotop voxels are within the range of the layer
        bool_z = xr.zeros_like(z)
        for i in range(z.z.shape[0]):
            bool_z[:, :, i] = np.where(
                (z[:, :, i] >= bot[:, :, lay].T) * (z[:, :, i] < top[:, :, lay].T), True, False)

        kh_geo = xr.where(bool_z, geotop_ds_raw['kh_from_litho'], np.nan)
        kh[:, :, lay] = kh_geo.mean(dim='z').T

    da_top = xr.DataArray(data=top, dims=('y', 'x', 'layer'),
                          coords={'y': geotop_ds_raw.y, 'x': geotop_ds_raw.x,
                                  'layer': geo_names})
    da_bot = xr.DataArray(data=bot, dims=('y', 'x', 'layer'),
                          coords={'y': geotop_ds_raw.y, 'x': geotop_ds_raw.x,
                                  'layer': geo_names})
    da_kh = xr.DataArray(data=kh, dims=('y', 'x', 'layer'),
                         coords={'y': geotop_ds_raw.y, 'x': geotop_ds_raw.x,
                                 'layer': geo_names})
    da_thick = xr.DataArray(data=thickness, dims=('y', 'x', 'layer'),
                            coords={'y': geotop_ds_raw.y, 'x': geotop_ds_raw.x,
                                    'layer': geo_names})

    geotop_ds_mod = xr.Dataset()

    geotop_ds_mod['top'] = da_top
    geotop_ds_mod['bot'] = da_bot
    geotop_ds_mod['kh'] = da_kh
    geotop_ds_mod['kv'] = geotop_ds_mod['kh'] * f_anisotropy
    geotop_ds_mod['thickness'] = da_thick

    return geotop_ds_mod



# -*- coding: utf-8 -*-
"""Created on Thu Jan  7 12:13:44 2021.

@author: oebbe
"""

import os

from . import mdims, mfpackages, read, util, visualise
from .version import __version__

NLMOD_DATADIR = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             '..', 'data')



# -*- coding: utf-8 -*-
"""Created on Fri Apr  2 15:08:50 2021.

@author: oebbe
"""
import logging

import numpy as np
import xarray as xr
from scipy import interpolate
from scipy.interpolate import griddata

from . import mgrid

logger = logging.getLogger(__name__)


def resample_dataarray2d_to_vertex_grid(da_in, gridprops=None,
                                        xyi=None, cid=None,
                                        method='nearest',
                                        **kwargs):
    """resample a 2d dataarray (xarray) from a structured grid to a new
    dataaraay of a vertex grid.

    Parameters
    ----------
    da_in : xarray.DataArray
        data array with dimensions (y, x). y and x are from the original
        grid
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen.
    xyi : numpy.ndarray, optional
        array with x and y coördinates of cell centers, shape(len(cid), 2). If
        xyi is None xyi is calculated from the gridproperties.
    cid : list or numpy.ndarray, optional
        list with cellids. If  cid is None cid is calculated from the
        gridproperties.
    method : str, optional
        type of interpolation used to resample. The default is 'nearest'.

    Returns
    -------
    da_out : xarray.DataArray
        data array with dimension (cid).
    """
    if (xyi is None) or (cid is None):
        xyi, cid = mgrid.get_xyi_cid(gridprops=gridprops,
                                     model_ds=kwargs.pop("model_ds", None))

    # get x and y values of all cells in dataarray
    mg = np.meshgrid(da_in.x.data, da_in.y.data)
    points = np.vstack((mg[0].ravel(), mg[1].ravel())).T

    # regrid
    arr_out = griddata(points, da_in.data.flatten(), xyi, method=method,
                       **kwargs)

    # new dataset
    da_out = xr.DataArray(arr_out, dims=('cid'),
                          coords={'cid': cid})

    return da_out


def resample_dataarray3d_to_vertex_grid(da_in, gridprops=None,
                                        xyi=None, cid=None,
                                        method='nearest'):
    """resample a dataarray (xarray) from a structured grid to a new dataaraay
    of a vertex grid.

    Parameters
    ----------
    da_in : xarray.DataArray
        data array with dimensions (layer, y, x). y and x are from the original
        grid
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen.
    xyi : numpy.ndarray, optional
        array with x and y coördinates of cell centers, shape(len(cid), 2). If
        xyi is None xyi is calculated from the gridproperties.
    cid : list or numpy.ndarray, optional
        list with cellids. If  cid is None cid is calculated from the
        gridproperties.
    method : str, optional
        type of interpolation used to resample. The default is 'nearest'.

    Returns
    -------
    da_out : xarray.DataArray
        data array with dimensions (layer,cid).
    """
    if (xyi is None) or (cid is None):
        xyi, cid = mgrid.get_xyi_cid(gridprops=gridprops)

    # get x and y values of all cells in dataarray
    mg = np.meshgrid(da_in.x.data, da_in.y.data)
    points = np.vstack((mg[0].ravel(), mg[1].ravel())).T

    layers = da_in.layer.data
    arr_out = np.zeros((len(layers), len(xyi)))
    for i, lay in enumerate(layers):

        ds_lay = da_in.sel(layer=lay)

        # regrid
        arr_out[i] = griddata(
            points, ds_lay.data.flatten(), xyi, method=method)

    # new dataset
    da_out = xr.DataArray(arr_out, dims=('layer', 'cid'),
                          coords={'cid': cid,
                                  'layer': layers})

    return da_out


def resample_dataset_to_vertex_grid(ds_in, gridprops,
                                    method='nearest'):
    """resample a dataset (xarray) from an structured grid to a new dataset
    from a vertex grid.

    Parameters
    ----------
    ds_in : xarray.Dataset
        dataset with dimensions (layer, y, x). y and x are from the original
        structured grid
    gridprops : dictionary
        dictionary with grid properties output from gridgen.
    method : str, optional
        type of interpolation used to resample. The default is 'nearest'.

    Returns
    -------
    ds_out : xarray.Dataset
        dataset with dimensions (layer, cid), cid are cell id's from the new
        grid.
    """

    assert isinstance(ds_in, xr.core.dataset.Dataset)

    xyi, cid = mgrid.get_xyi_cid(gridprops)

    ds_out = xr.Dataset(coords={'cid': cid,
                                'layer': ds_in.layer.data})

    # add x and y coordinates
    ds_out['x'] = xr.DataArray(xyi[:, 0], dims=('cid'),
                               coords={'cid': cid})
    ds_out['y'] = xr.DataArray(xyi[:, 0], dims=('cid'),
                               coords={'cid': cid})

    # add other variables
    for data_var in ds_in.data_vars:
        if ds_in[data_var].dims == ('layer', 'y', 'x'):
            data_arr = resample_dataarray3d_to_vertex_grid(ds_in[data_var],
                                                           xyi=xyi, cid=cid,
                                                           method=method)
        elif ds_in[data_var].dims == ('y', 'x'):
            data_arr = resample_dataarray2d_to_vertex_grid(ds_in[data_var],
                                                           xyi=xyi, cid=cid,
                                                           method=method)

        elif ds_in[data_var].dims == ('layer') or ds_in[data_var].dims == ('layer',):
            data_arr = ds_in[data_var]

        else:
            logger.warning(
                f'did not resample data array {data_var} because conversion with dimensions {ds_in[data_var].dims} is not (yet) supported')
            continue

        ds_out[data_var] = data_arr

    return ds_out


def resample_dataarray2d_to_structured_grid(da_in, extent=None,
                                            delr=None, delc=None,
                                            xmid=None, ymid=None,
                                            kind='linear', nan_factor=0.01,
                                            **kwargs):
    """resample a dataarray (xarray) from a structured grid to a new dataaraay
    from a different structured grid.

    Parameters
    ----------
    da_in : xarray.DataArray
        data array with dimensions (y, x). y and x are from the original
        grid
    extent : list, tuple or np.array, optional
        extent (xmin, xmax, ymin, ymax) of the desired grid, if not defined
        xmid and ymid are used
    delr : int or float, optional
        cell size along rows of the desired grid, if not defined xmid and
        ymid are used
    delc : int or float, optional
        cell size along columns of the desired grid, if not defined xmid and
        ymid are used
    xmid : np.array, optional
        x coördinates of the cell centers of the desired grid shape(ncol), if
        not defined xmid and ymid are calculated from the extent, delr and delc.
    ymid : np.array, optional
        y coördinates of the cell centers of the desired grid shape(nrow), if
        not defined xmid and ymid are calculated from the extent, delr and delc.
    kind : str, optional
        type of interpolation used to resample. The default is 'linear'.
    nan_factor : float, optional
        the nan values in the original raster are filled with zeros before
        interpolation because the interp2d function cannot handle nan values
        very well. Therefore an extra interpolation is done to determine how
        much these nan values have influenced the new raster values. If the
        the interpolated value is influenced more than this factor by a nan
        value. The value in the interpolated raster is set to nan.
        See also: https://stackoverflow.com/questions/51474792/2d-interpolation-with-nan-values-in-python

    Returns
    -------
    ds_out : xarray.DataArray
        data array with dimensions (y, x). y and x are from the new grid.
    """

    assert isinstance(
        da_in, xr.core.dataarray.DataArray), f'expected type xr.core.dataarray.DataArray got {type(da_in)} instead'

    if xmid is None:
        xmid, ymid = mgrid.get_xy_mid_structured(extent, delr, delc)

    # check if ymid is in descending order
    assert np.array_equal(ymid, np.sort(
        ymid)[::-1]), 'ymid should be in descending order'

    # check for nan values
    if (da_in.isnull().sum() > 0) and (kind == "linear"):
        arr_out = resample_2d_struc_da_nan_linear(da_in, xmid, ymid,
                                                  nan_factor, **kwargs)
    # faster for linear
    elif kind in ["linear", 'cubic']:
        # no need to fill nan values
        f = interpolate.interp2d(da_in.x.data, da_in.y.data,
                                 da_in.data, kind='linear', **kwargs)
        # for some reason interp2d flips the y-values
        arr_out = f(xmid, ymid)[::-1]
    elif kind == 'nearest':
        xydata = np.vstack([v.ravel() for v in
                            np.meshgrid(da_in.x.data, da_in.y.data)]).T
        xyi = np.vstack([v.ravel() for v in np.meshgrid(xmid, ymid)]).T
        fi = griddata(xydata, da_in.data.ravel(), xyi, method=kind,
                      **kwargs)
        arr_out = fi.reshape(ymid.shape[0], xmid.shape[0])
    else:
        raise ValueError(f'unexpected value for "kind": {kind}')

    # new dataset
    da_out = xr.DataArray(arr_out, dims=('y', 'x'),
                          coords={'x': xmid,
                                  'y': ymid})

    return da_out


def resample_dataarray3d_to_structured_grid(da_in, extent=None,
                                            delr=None, delc=None,
                                            xmid=None, ymid=None,
                                            kind='linear', nan_factor=0.01,
                                            **kwargs):
    """resample a dataarray (xarray) from a structured grid to a new dataaraay
    from a different structured grid.

    Parameters
    ----------
    da_in : xarray.DataArray
        data array with dimensions (layer, y, x). y and x are from the original
        grid
    extent : list, tuple or np.array, optional
        extent (xmin, xmax, ymin, ymax) of the desired grid, if not defined
        xmid and ymid are used
    delr : int or float, optional
        cell size along rows of the desired grid, if not defined xmid and
        ymid are used
    delc : int or float, optional
        cell size along columns of the desired grid, if not defined xmid and
        ymid are used
    xmid : np.array, optional
        x coördinates of the cell centers of the desired grid shape(ncol), if
        not defined xmid and ymid are calculated from the extent, delr and delc.
    ymid : np.array, optional
        y coördinates of the cell centers of the desired grid shape(nrow), if
        not defined xmid and ymid are calculated from the extent, delr and delc.
    kind : str, optional
        type of interpolation used to resample. The default is 'linear'.
    nan_factor : float, optional
        the nan values in the original raster are filled with zeros before
        interpolation because the interp2d function cannot handle nan values
        very well. Therefore an extra interpolation is done to determine how
        much these nan values have influenced the new raster values. If the
        the interpolated value is influenced more than this factor by a nan
        value. The value in the interpolated raster is set to nan.
        See also: https://stackoverflow.com/questions/51474792/2d-interpolation-with-nan-values-in-python

    Returns
    -------
    ds_out : xarray.DataArray
        data array with dimensions (layer, y, x). y and x are from the new
        grid.
    """

    assert isinstance(
        da_in, xr.core.dataarray.DataArray), f'expected type xr.core.dataarray.DataArray got {type(da_in)} instead'

    # check if ymid is in descending order
    assert np.array_equal(ymid, np.sort(
        ymid)[::-1]), 'ymid should be in descending order'

    if xmid is None:
        xmid, ymid = mgrid.get_xy_mid_structured(extent, delr, delc)

    layers = da_in.layer.data
    arr_out = np.zeros((len(layers), len(ymid), len(xmid)))
    for i, lay in enumerate(layers):

        ds_lay = da_in.sel(layer=lay)
        # check for nan values
        if (ds_lay.isnull().sum() > 0) and (kind == "linear"):
            arr_out[i] = resample_2d_struc_da_nan_linear(ds_lay, xmid, ymid,
                                                         nan_factor, **kwargs)
        # faster for linear
        elif kind in ['linear', 'cubic']:
            # no need to fill nan values
            f = interpolate.interp2d(ds_lay.x.data, ds_lay.y.data,
                                     ds_lay.data, kind='linear', **kwargs)
            # for some reason interp2d flips the y-values
            arr_out[i] = f(xmid, ymid)[::-1]
        elif kind == 'nearest':
            xydata = np.vstack([v.ravel() for v in
                                np.meshgrid(ds_lay.x.data, ds_lay.y.data)]).T
            xyi = np.vstack([v.ravel() for v in np.meshgrid(xmid, ymid)]).T
            fi = griddata(xydata, ds_lay.data.ravel(), xyi, method=kind,
                          **kwargs)
            arr_out[i] = fi.reshape(ymid.shape[0], xmid.shape[0])
        else:
            raise ValueError(f'unexpected value for "kind": {kind}')

    # new dataset
    da_out = xr.DataArray(arr_out, dims=('layer', 'y', 'x'),
                          coords={'x': xmid,
                                  'y': ymid,
                                  'layer': layers})

    return da_out


def resample_2d_struc_da_nan_linear(da_in, new_x, new_y,
                                    nan_factor=0.01, **kwargs):
    """resample a structured, 2d data-array with nan values onto a new grid.

    Parameters
    ----------
    da_in : xarray DataArray
        dataset you want to project on a new grid
    new_x : numpy array
        x coördinates of the new grid
    new_y : numpy array
        y coördaintes of the new grid
    nan_factor : float, optional
        the nan values in the original raster are filled with zeros before
        interpolation because the interp2d function cannot handle nan values
        very well. Therefore an extra interpolation is done to determine how
        much these nan values have influenced the new raster values. If the
        the interpolated value is influenced more than this factor by a nan
        value. The value in the interpolated raster is set to nan.
        See also: https://stackoverflow.com/questions/51474792/2d-interpolation-with-nan-values-in-python

    Returns
    -------
    arr_out : numpy array
        resampled array
    """
    nan_map = np.where(da_in.isnull().data, 1, 0)
    fill_map = np.where(da_in.isnull().data, 0, da_in.data)
    f = interpolate.interp2d(da_in.x.data, da_in.y.data,
                             fill_map, kind='linear', **kwargs)
    f_nan = interpolate.interp2d(da_in.x.data, da_in.y.data,
                                 nan_map, kind='linear')
    arr_out_raw = f(new_x, new_y)
    nan_new = f_nan(new_x, new_y)
    arr_out_raw[nan_new > nan_factor] = np.nan

    # for some reason interp2d flips the y-values
    arr_out = arr_out_raw[::-1]

    return arr_out


def resample_dataset_to_structured_grid(ds_in, extent, delr, delc, kind='linear'):
    """Resample a dataset (xarray) from a structured grid to a new dataset from
    a different structured grid.

    Parameters
    ----------
    ds_in : xarray.Dataset
        dataset with dimensions (layer, y, x). y and x are from the original
        grid
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax) of the desired grid.
    delr : int or float
        cell size along rows of the desired grid (dx).
    delc : int or float
        cell size along columns of the desired grid (dy).
    kind : str, optional
        type of interpolation used to resample. The default is 'linear'.

    Returns
    -------
    ds_out : xarray.Dataset
        dataset with dimensions (layer, y, x). y and x are from the new
        grid.
    """

    assert isinstance(ds_in, xr.core.dataset.Dataset)

    xmid, ymid = mgrid.get_xy_mid_structured(extent, delr, delc)

    ds_out = xr.Dataset(coords={'y': ymid,
                                'x': xmid,
                                'layer': ds_in.layer.data})
    for data_var in ds_in.data_vars:
        data_arr = resample_dataarray3d_to_structured_grid(ds_in[data_var],
                                                           xmid=xmid,
                                                           ymid=ymid,
                                                           kind=kind)
        ds_out[data_var] = data_arr

    return ds_out


def get_resampled_ml_layer_ds_vertex(raw_ds=None,
                                     extent=None,
                                     gridprops=None):
    """Project model layer dataset on a vertex model grid.

    Parameters
    ----------
    raw_ds : xr.Dataset, optional
        raw model layer dataset. The default is None.
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax) of the desired grid.
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen. Used as the
        definition of the vertex grid.

    Returns
    -------
    ml_layer_ds : xr.dataset
        model layer dataset projected onto the modelgrid.
    """

    logger.info('resample model layer data to vertex modelgrid')
    ml_layer_ds = resample_dataset_to_vertex_grid(
        raw_ds, gridprops)
    ml_layer_ds['x'] = xr.DataArray([r[1] for r in gridprops['cell2d']],
                                    dims=('cid'),
                                    coords={'cid': ml_layer_ds.cid.data})

    ml_layer_ds['y'] = xr.DataArray([r[2] for r in gridprops['cell2d']],
                                    dims=('cid'),
                                    coords={'cid': ml_layer_ds.cid.data})
    ml_layer_ds.attrs['gridtype'] = 'vertex'
    ml_layer_ds.attrs['delr'] = raw_ds.delr
    ml_layer_ds.attrs['delc'] = raw_ds.delc
    ml_layer_ds.attrs['extent'] = extent

    return ml_layer_ds


def fillnan_dataarray_structured_grid(xar_in, method='nearest'):
    """fill not-a-number values in a structured grid, DataArray.

    The fill values are determined using the 'nearest' method of the
    scipy.interpolate.griddata function


    Parameters
    ----------
    xar_in : xarray DataArray
        DataArray with nan values. DataArray should have 2 dimensions
        (y and x).
    method : str, optional
        method used in scipy.interpolate.griddata to resample, default is
        nearest.

    Returns
    -------
    xar_out : xarray DataArray
        DataArray without nan values. DataArray has 2 dimensions
        (y and x)

    Notes
    -----
    can be slow if the xar_in is a large raster
    """
    # check dimensions
    if xar_in.dims != ('y', 'x'):
        raise ValueError(
            f"expected dataarray with dimensions ('y' and 'x'), got dimensions -> {xar_in.dims}")

    # get list of coordinates from all points in raster
    mg = np.meshgrid(xar_in.x.data, xar_in.y.data)
    points_all = np.vstack((mg[0].ravel(), mg[1].ravel())).T

    # get all values in DataArray
    values_all = xar_in.data.flatten()

    # get 1d arrays with only values where DataArray is not nan
    mask1 = ~np.isnan(values_all)
    points_in = points_all[np.where(mask1)[0]]
    values_in = values_all[np.where(mask1)[0]]

    # get value for all nan values
    values_out = griddata(points_in, values_in, points_all, method=method)
    arr_out = values_out.reshape(xar_in.shape)

    # create DataArray without nan values
    xar_out = xr.DataArray(arr_out, dims=('y', 'x'),
                           coords={'x': xar_in.x.data,
                                   'y': xar_in.y.data})

    return xar_out


def fillnan_dataarray_vertex_grid(xar_in, gridprops=None,
                                  xyi=None, cid=None,
                                  method='nearest'):
    """fill not-a-number values in a vertex grid, DataArray.

    The fill values are determined using the 'nearest' method of the
    scipy.interpolate.griddata function

    Parameters
    ----------
    xar_in : xr.DataArray
        data array with nan values. Shape is (cid)
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen.
    xyi : numpy.ndarray
        array with x and y coördinates of cell centers, shape(len(cid), 2).
    cid : list
        list with cellids.
    method : str, optional
        method used in scipy.interpolate.griddata to resample, default is
        nearest.

    Returns
    -------
    xar_out : xr.DataArray
        data array with nan values. Shape is (cid)

    Notes
    -----
    can be slow if the xar_in is a large raster
    """

    # get list of coordinates from all points in raster
    if (xyi is None) or (cid is None):
        xyi, cid = mgrid.get_xyi_cid(gridprops)

    # fill nan values in DataArray
    values_all = xar_in.data

    # get 1d arrays with only values where DataArray is not nan
    mask1 = ~np.isnan(values_all)
    xyi_in = xyi[mask1]
    values_in = values_all[mask1]

    # get value for all nan values
    values_out = griddata(xyi_in, values_in, xyi, method=method)

    # create DataArray without nan values
    xar_out = xr.DataArray(values_out, dims=('cid'),
                           coords={'cid': xar_in.cid.data})

    return xar_out


def resample_vertex_2d_da_to_struc_2d_da(da_in, model_ds=None,
                                         xmid=None, ymid=None,
                                         cellsize=25,
                                         method='nearest'):
    """resample a 2d dataarray (xarray) from a vertex grid to a new dataaraay
    from a structured grid.

    Parameters
    ----------
    da_in : xarray.DataArray
        data array with dimensions ('cid').
    model_ds : xarray.DataArray
        model dataset with 'x' and 'y' data variables.
    cellsize : int or float, optional
        required cell size of structured grid. The default is 25.
    method : str, optional
        method used for resampling. The default is 'nearest'.

    Returns
    -------
    da_out : xarray.DataArray
        data array with dimensions ('y', 'x').
    """
    if xmid is None or ymid is None:
        xmid = model_ds.x.values
        ymid = model_ds.y.values

    points_vertex = np.array([xmid, ymid]).T
    modelgrid_x = np.arange(xmid.min(),
                            xmid.max(),
                            cellsize)
    modelgrid_y = np.arange(ymid.max(),
                            ymid.min() - cellsize,
                            -cellsize)
    mg = np.meshgrid(modelgrid_x, modelgrid_y)
    points = np.vstack((mg[0].ravel(), mg[1].ravel())).T

    arr_out_1d = griddata(points_vertex, da_in.values, points, method=method)
    arr_out2d = arr_out_1d.reshape(len(modelgrid_y),
                                   len(modelgrid_x))

    da_out = xr.DataArray(arr_out2d,
                          dims=('y', 'x'),
                          coords={'y': modelgrid_y,
                                  'x': modelgrid_x})

    return da_out



# -*- coding: utf-8 -*-
"""Created on Thu Jan  7 17:20:34 2021.

@author: oebbe
"""
import logging
import numbers
import os
import sys

import flopy
import numpy as np
import xarray as xr

from .. import mdims
from . import recharge

logger = logging.getLogger(__name__)


def sim_tdis_gwf_ims_from_model_ds(model_ds,
                                   complexity='MODERATE',
                                   exe_name=None):
    """create sim, tdis, gwf and ims package from the model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data. Should have the dimension 'time' and the
        attributes: model_name, mfversion, model_ws, time_units, start_time,
        perlen, nstp, tsmult
    exe_name: str, optional
        path to modflow executable, default is None, which assumes binaries
        are available in nlmod/bin directory. Binaries can be downloaded
        using `nlmod.util.download_mfbinaries()`.

    Returns
    -------
    sim : flopy MFSimulation
        simulation object.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    """

    # start creating model
    logger.info('creating modflow SIM, TDIS, GWF and IMS')

    if exe_name is None:
        exe_name = os.path.join(os.path.dirname(__file__),
                                '..', '..', 'bin', model_ds.mfversion)
        if sys.platform.startswith('win'):
            exe_name += ".exe"

    # Create the Flopy simulation object
    sim = flopy.mf6.MFSimulation(sim_name=model_ds.model_name,
                                 exe_name=exe_name,
                                 version=model_ds.mfversion,
                                 sim_ws=model_ds.model_ws)

    tdis_perioddata = get_tdis_perioddata(model_ds)

    # Create the Flopy temporal discretization object
    flopy.mf6.modflow.mftdis.ModflowTdis(sim,
                                         pname='tdis',
                                         time_units=model_ds.time_units,
                                         nper=len(model_ds.time),
                                         start_date_time=model_ds.start_time,
                                         perioddata=tdis_perioddata)

    # Create the Flopy groundwater flow (gwf) model object
    model_nam_file = '{}.nam'.format(model_ds.model_name)
    gwf = flopy.mf6.ModflowGwf(sim, modelname=model_ds.model_name,
                               model_nam_file=model_nam_file)

    # Create the Flopy iterative model solver (ims) Package object
    flopy.mf6.modflow.mfims.ModflowIms(sim, pname='ims',
                                       print_option='summary',
                                       complexity=complexity)

    return sim, gwf


def dis_from_model_ds(model_ds, gwf, length_units='METERS',
                      angrot=0):
    """get discretisation package from the model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    length_units : str, optional
        length unit. The default is 'METERS'.
    angrot : int or float, optional
        rotation angle. The default is 0.

    Returns
    -------
    dis : TYPE
        discretisation package.
    """

    if model_ds.gridtype != 'structured':
        raise ValueError(
            f'cannot create dis package for gridtype -> {model_ds.gridtype}')

    # check attributes
    for att in ['delr', 'delc']:
        if isinstance(model_ds.attrs[att], np.float32):
            model_ds.attrs[att] = float(model_ds.attrs[att])

    dis = flopy.mf6.ModflowGwfdis(gwf,
                                  pname='dis',
                                  length_units=length_units,
                                  xorigin=model_ds.extent[0],
                                  yorigin=model_ds.extent[2],
                                  angrot=angrot,
                                  nlay=model_ds.dims['layer'],
                                  nrow=model_ds.dims['y'],
                                  ncol=model_ds.dims['x'],
                                  delr=model_ds.delr,
                                  delc=model_ds.delc,
                                  top=model_ds['top'].data,
                                  botm=model_ds['bot'].data,
                                  idomain=model_ds['idomain'].data,
                                  filename=f'{model_ds.model_name}.dis')

    return dis


def disv_from_model_ds(model_ds, gwf, gridprops,
                       length_units='METERS',
                       angrot=0):
    """get discretisation vertices package from the model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    gridprops : dictionary
        dictionary with grid properties output from gridgen.
    length_units : str, optional
        length unit. The default is 'METERS'.
    angrot : int or float, optional
        rotation angle. The default is 0.

    Returns
    -------
    disv : flopy ModflowGwfdisv
        disv package
    """

    disv = flopy.mf6.ModflowGwfdisv(gwf,
                                    idomain=model_ds['idomain'].data,
                                    xorigin=model_ds.extent[0],
                                    yorigin=model_ds.extent[2],
                                    length_units=length_units,
                                    angrot=angrot,
                                    **gridprops)

    return disv


def npf_from_model_ds(model_ds, gwf, icelltype=0,
                      save_flows=False,
                      **kwargs):
    """get node property flow package from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    icelltype : int, optional
        celltype. The default is 0.
    save_flows : bool, optional
        value is passed to flopy.mf6.ModflowGwfnpf() to determine if cell by
        cell flows should be saved to the cbb file. Default is False

    Raises
    ------
    NotImplementedError
        only icelltype 0 is implemented.

    Returns
    -------
    npf : flopy ModflowGwfnpf
        npf package.
    """

    npf = flopy.mf6.ModflowGwfnpf(gwf,
                                  pname='npf',
                                  icelltype=icelltype,
                                  k=model_ds['kh'].data,
                                  k33=model_ds['kv'].data,
                                  save_flows=save_flows,
                                  **kwargs)

    return npf


def ghb_from_model_ds(model_ds, gwf, da_name):
    """get general head boundary from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    da_name : str
        name of the ghb files in the model dataset.

    Raises
    ------
    ValueError
        raised if gridtype is not structured or vertex.

    Returns
    -------
    ghb : flopy ModflowGwfghb
        ghb package
    """

    if model_ds.gridtype == 'structured':
        ghb_rec = mdims.data_array_2d_to_rec_list(model_ds,
                                                  model_ds[f'{da_name}_cond'] != 0,
                                                  col1=f'{da_name}_peil',
                                                  col2=f'{da_name}_cond',
                                                  first_active_layer=True,
                                                  only_active_cells=False,
                                                  layer=0)
    elif model_ds.gridtype == 'vertex':
        ghb_rec = mdims.data_array_1d_vertex_to_rec_list(model_ds,
                                                         model_ds[f'{da_name}_cond'] != 0,
                                                         col1=f'{da_name}_peil',
                                                         col2=f'{da_name}_cond',
                                                         first_active_layer=True,
                                                         only_active_cells=False,
                                                         layer=0)
    else:
        raise ValueError(f'did not recognise gridtype {model_ds.gridtype}')

    if len(ghb_rec) > 0:
        ghb = flopy.mf6.ModflowGwfghb(gwf, print_input=True,
                                      maxbound=len(ghb_rec),
                                      stress_period_data=ghb_rec,
                                      save_flows=True)
        return ghb

    else:
        print('no ghb cells added')

        return None


def ic_from_model_ds(model_ds, gwf,
                     starting_head='starting_head'):
    """get initial condictions package from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    starting_head : str, float or int, optional
        if type is int or float this is the starting head for all cells
        If the type is str the data variable from model_ds is used as starting
        head. The default is 'starting_head'.

    Returns
    -------
    ic : flopy ModflowGwfic
        ic package
    """
    if isinstance(starting_head, str):
        pass
    elif isinstance(starting_head, numbers.Number):
        model_ds['starting_head'] = starting_head * \
            xr.ones_like(model_ds['idomain'])
        model_ds['starting_head'].attrs['units'] = 'mNAP'
        starting_head = 'starting_head'

    ic = flopy.mf6.ModflowGwfic(gwf, pname='ic',
                                strt=model_ds[starting_head].data)

    return ic


def sto_from_model_ds(model_ds, gwf,
                      sy=0.2, ss=0.000001,
                      iconvert=1, save_flows=False):
    """get storage package from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    sy : float, optional
        DESCRIPTION. The default is 0.2.
    ss : float, optional
        specific storage. The default is 0.000001.
    iconvert : int, optional
        DESCRIPTION. The default is 1.
    save_flows : bool, optional
        value is passed to flopy.mf6.ModflowGwfsto() to determine if flows
        should be saved to the cbb file. Default is False

    Returns
    -------
    sto : flopy ModflowGwfsto
        sto package
    """

    if model_ds.steady_state:
        return None
    else:
        if model_ds.steady_start:
            sts_spd = {0: True}
            trn_spd = {1: True}
        else:
            sts_spd = None
            trn_spd = {0: True}

        sto = flopy.mf6.ModflowGwfsto(gwf, pname='sto',
                                      save_flows=save_flows,
                                      iconvert=iconvert,
                                      ss=ss, sy=sy, steady_state=sts_spd,
                                      transient=trn_spd)
        return sto


def chd_from_model_ds(model_ds, gwf, chd='chd',
                      head='starting_head'):
    """get constant head boundary at the model's edges from the model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    chd : str, optional
        name of data variable in model_ds that is 1 for cells with a constant
        head and zero for all other cells. The default is 'chd'.
    head : str, optional
        name of data variable in model_ds that is used as the head in the chd
        cells. The default is 'starting_head'.

    Returns
    -------
    chd : flopy ModflowGwfchd
        chd package
    """
    # get the stress_period_data
    if model_ds.gridtype == 'structured':
        chd_rec = mdims.data_array_3d_to_rec_list(model_ds,
                                                  model_ds[chd] != 0,
                                                  col1=head)
    elif model_ds.gridtype == 'vertex':
        cellids = np.where(model_ds[chd])
        chd_rec = list(zip(zip(cellids[0],
                               cellids[1]),
                           [1.0] * len(cellids[0])))

    chd = flopy.mf6.ModflowGwfchd(gwf, pname=chd,
                                  maxbound=len(chd_rec),
                                  stress_period_data=chd_rec,
                                  save_flows=True)

    return chd


def surface_drain_from_model_ds(model_ds, gwf, surface_drn_cond=1000):
    """get surface level drain (maaivelddrainage in Dutch) from the model
    dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.
    surface_drn_cond : int or float, optional
        conductivity of the surface drain. The default is 1000.

    Returns
    -------
    drn : flopy ModflowGwfdrn
        drn package
    """

    model_ds.attrs['surface_drn_cond'] = surface_drn_cond
    mask = model_ds['ahn'].notnull()
    if model_ds.gridtype == 'structured':
        drn_rec = mdims.data_array_2d_to_rec_list(model_ds, mask, col1='ahn',
                                                  first_active_layer=True,
                                                  only_active_cells=False,
                                                  col2=model_ds.surface_drn_cond)
    elif model_ds.gridtype == 'vertex':
        drn_rec = mdims.data_array_1d_vertex_to_rec_list(model_ds, mask,
                                                         col1='ahn',
                                                         col2=model_ds.surface_drn_cond,
                                                         first_active_layer=True,
                                                         only_active_cells=False)

    drn = flopy.mf6.ModflowGwfdrn(gwf, print_input=True,
                                  maxbound=len(drn_rec),
                                  stress_period_data={0: drn_rec},
                                  save_flows=True)

    return drn


def rch_from_model_ds(model_ds, gwf):
    """get recharge package from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.

    Returns
    -------
    rch : flopy ModflowGwfrch
        rch package
    """

    # create recharge package
    rch = recharge.model_datasets_to_rch(gwf, model_ds)

    return rch


def oc_from_model_ds(model_ds, gwf, save_budget=True,
                     print_head=True):
    """get output control package from model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    gwf : flopy ModflowGwf
        groundwaterflow object.

    Returns
    -------
    oc : flopy ModflowGwfoc
        oc package
    """
    # Create the output control package
    headfile = '{}.hds'.format(model_ds.model_name)
    head_filerecord = [headfile]
    budgetfile = '{}.cbb'.format(model_ds.model_name)
    budget_filerecord = [budgetfile]
    saverecord = [('HEAD', 'LAST')]
    if save_budget:
        saverecord.append(('BUDGET', 'ALL'))
    if print_head:
        printrecord = [('HEAD', 'LAST')]
    else:
        printrecord = None

    oc = flopy.mf6.ModflowGwfoc(gwf, pname='oc',
                                saverecord=saverecord,
                                head_filerecord=head_filerecord,
                                budget_filerecord=budget_filerecord,
                                printrecord=printrecord)

    return oc


def get_tdis_perioddata(model_ds):
    """Get tdis_perioddata from model_ds.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with time variant model data

    Returns
    -------
    tdis_perioddata : [perlen, nstp, tsmult]
        - perlen (double) is the length of a stress period.
        - nstp (integer) is the number of time steps in a stress period.
        - tsmult (double) is the multiplier for the length of successive time
          steps. The length of a time step is calculated by multiplying the
          length of the previous time step by TSMULT. The length of the first
          time step, :math:`\\Delta t_1`, is related to PERLEN, NSTP, and
          TSMULT by the relation :math:`\\Delta t_1= perlen \frac{tsmult -
          1}{tsmult^{nstp}-1}`.
    """
    perlen = model_ds.perlen
    if isinstance(perlen, numbers.Number):
        tdis_perioddata = [(float(perlen), model_ds.nstp,
                            model_ds.tsmult)] * int(model_ds.nper)
    elif isinstance(perlen, (list, tuple, np.ndarray)):
        if model_ds.steady_start:
            assert len(perlen) == model_ds.dims['time']
        else:
            assert len(perlen) == model_ds.dims['time']
        tdis_perioddata = [(p, model_ds.nstp, model_ds.tsmult) for p in perlen]
    else:
        raise TypeError('did not recognise perlen type')

    # netcdf does not support multi-dimensional array attributes
    #model_ds.attrs['tdis_perioddata'] = tdis_perioddata

    return tdis_perioddata
